import math

from .entropy import shannon_entropy


def binary_derivative(string):
    """
    Calculates the binary derivative of the given string, according to:

    .. code-block:: text

        Nathanson, M. B. (1971). Derivatives of binary sequences.
        SIAM Journal on Applied Mathematics, 21(3), 407-412

    The derivative of a binary string is simply the bitwise exclusive OR between a binary digit in the string and
    its successor, if a successor exists. For example, the derivative of the binary string '01010101' is the binary
    string '1111111'.

    :param string: a binary string, such as '110011'

    :return: a binary string representing the binary derivative of the given string
    """
    result = []
    for i, d in enumerate(string):
        if i - 1 == len(string) - 2:
            break
        result.append(int(string[i]) ^ int(string[i + 1]))
    return ''.join([str(x) for x in result])


def bien(string):
    """
    Calculate the BiEntropy of the given binary string, according to:

    .. code-block:: text

        Croll, G. J. (2013). BiEntropy-The Approximate Entropy of a Finite Binary String.
        arXiv preprint arXiv:1305.0954.

    The BiEntropy can be used to compare two binary strings in terms of the relative order and disorder of all the
    digits. It makes use of a weighted average of the Shannon entropies of all but the last binary derivative of the
    given string. This version of BiEntropy is suitable for strings with length <= 32.

    :param string: a binary string, such as '110011'

    :return: a real number representing the BiEntropy of the given string
    """
    tot = 0.0
    n = len(string)
    for k in range(n - 1):
        tot += shannon_entropy(string) * 2**k
        string = binary_derivative(string)
    return (1 / (2**(n - 1) - 1)) * tot


def tbien(string):
    """
    Calculates the logarithmic weighting BiEntropy of the given string, according to:

    .. code-block:: text

        Croll, G. J. (2013). BiEntropy-The Approximate Entropy of a Finite Binary String.
        arXiv preprint arXiv:1305.0954.

    The Logarithmic BiEntropy can be used to compare two binary strings in terms of the relative order and disorder of
    all the digits. It makes use of a logarithmic weighted average of the Shannon entropies of all but the last binary
    derivative of the given string. This version of BiEntropy is suitable for strings with length > 32.

    :param string: a binary string, such as '110011'

    :return: a real number representing the logarithmic weighting BiEntropy of the given string
    """
    tot = 0.0
    tot_log = 0.0
    n = len(string)
    for k in range(n - 1):
        lg = math.log(k + 2, 2.0)
        tot += shannon_entropy(string) * lg
        tot_log += lg
        string = binary_derivative(string)
    return (1 / tot_log) * tot


def cyclic_binary_derivative(string):
    """
    Calculates the cyclic binary derivative, which is the "binary string of length n formed by XORing adjacent pairs of 
    digits including the last and the first." See:

    .. code-block:: text

        Croll, G. J. (2018). The BiEntropy of Some Knots on the Simple Cubic Lattice.
        arXiv preprint arXiv:1802.03772.

    The cyclic derivative of a binary string is simply the bitwise exclusive OR between a binary digit in the string and
    its successor, where the successor of the last digit in the string is the first digit in the string. For example,
    the cyclic derivative of the binary string '01010101' is the binary string '11111111'.

    :param string: a binary string, such as '110011'

    :return: a binary string representing the cyclic binary derivative of the given string
    """
    result = []
    for i, d in enumerate(string):
        s = string[i]
        if i == len(string) - 1:
            next_s = string[0]
        else:
            next_s = string[i + 1]
        result.append(int(s) ^ int(next_s))
    return ''.join([str(x) for x in result])


def ktbien(string):
    """
    Calculates the knot logarithmic weighting BiEntropy of the given string, according to:

    .. code-block:: text

        Croll, G. J. (2018). The BiEntropy of Some Knots on the Simple Cubic Lattice.
        arXiv preprint arXiv:1802.03772.

    The Logarithmic Knot BiEntropy can be used to compare two binary strings in terms of the relative order and disorder
    of all the digits. It makes use of a logarithmic weighted average of the Shannon entropies of all but the last
    cyclic binary derivative of the given string. This version of BiEntropy is suitable for strings with length > 32.

    :param string: a binary string, such as '110011'

    :return: a real number representing the knot logarithmic weighting BiEntropy of the given string
    """
    tot = 0.0
    tot_log = 0.0
    n = len(string)
    for k in range(n - 1):
        lg = math.log(k + 2, 2.0)
        tot += shannon_entropy(string) * lg
        tot_log += lg
        string = cyclic_binary_derivative(string)
    return (1 / tot_log) * tot



import numpy as np


def apen(sequence, m=1, r=0):
    """
    Calculates the Approximate Entropy, or ApEn, of the given sequence, as described in:

    .. code-block:: text

      Pincus, S. M.; Gladstone, I. M.; Ehrenkranz, R. A. (1991).
      "A Regularity Statistic For Medical Data Analysis".
      Journal of Clinical Monitoring and Computing. 7 (4): 335-345.

    The implementation here is based on the Python implementation described in:
    https://en.wikipedia.org/wiki/Approximate_entropy

    :param sequence: a string of whole numbers, such as '012301', or an array of whole numbers, such as [0,1,2,3,0,1],
                     or a numpy array of whole numbers

    :param m: the length of compared runs of data

    :param r: a filtering level

    :return: a real number, representing the approximate entropy (ApEn) for the given sequence
    """
    if type(sequence) is str:
        U = np.array([int(x) for x in sequence])
    elif type(sequence) is list:
        U = np.array(sequence)
    elif type(sequence) is np.ndarray:
        U = sequence
    else:
        raise TypeError("unsupported sequence type: %s" % type(sequence))

    N = len(U)

    def maximum_distance(x_i, x_j):
        return max([abs(ua - va) for ua, va in zip(x_i, x_j)])

    def phi(m):
        x = [[U[j] for j in range(i, i + m - 1 + 1)] for i in range(N - m + 1)]
        C = [len([1 for x_j in x if maximum_distance(x_i, x_j) <= r]) / (N - m + 1.0) for x_i in x]
        return (1 / (N - m + 1.0)) * sum(np.log(C))

    return abs(phi(m + 1) - phi(m))



from .ca_functions import BaseRule


class CTRBLRule(BaseRule):
    """
    A rule that operates on von Neumann neighbourhoods, taking into account the states of a cell's
    neighbours at the top, right, bottom and left positions. Only supports 2D automata with periodic boundaries and a
    radius of 1.
    """
    def __init__(self, rule_table, add_rotations=False):
        """
        Creates a CTRBLRule.

        :param rule_table: a dictionary with keys being a 5-tuple representing the states of the CTRBL cells, and values
                           being a single value representing the image state (i.e. the state of the Center cell in
                           the next timestep); all combinations of states must exist, otherwise, if the combination of
                           states does not exist in the rule table, an exception will be raised

        :param add_rotations: whether rotations in the rule table are implied, and should be included (default is False)
        """
        self._rule_table = self._init_rule_table(rule_table, add_rotations)

    def __call__(self, n, c, t):
        """
        The CTRBL rule to apply.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        current_activity = n[1][1]
        top = n[0][1]
        right = n[1][2]
        bottom = n[2][1]
        left = n[1][0]
        key = (current_activity, top, right, bottom, left)
        if key not in self._rule_table:
            raise ValueError("neighbourhood state (%s, %s, %s, %s, %s) not in rule table" % key)
        return self._rule_table[key]

    @property
    def rule_table(self):
        """
        The rule table for this CTRBL rule.

        :return: the rule table
        """
        return self._rule_table

    @staticmethod
    def _init_rule_table(rule_table, add_rotations):
        new_rule_table = {}
        for rule, image in rule_table.items():
            new_rule_table[rule] = image
            if add_rotations:
                r = list(rule)
                for _ in range(3):
                    r.insert(1, r.pop(4))
                    new_rule_table[tuple(r)] = image
        return new_rule_table



import numpy as np
from .ctrbl_rule import CTRBLRule


class Evoloop(CTRBLRule):
    """
    An implementation of H. Sayama's Evoloop. For more information, see:

    .. code-block:: text

       Sayama, H. (1998). Constructing evolutionary systems on a simple deterministic cellular automata space.
       PhD, University of Tokyo, Department of Information Science.
    """
    def __init__(self):
        """
        Create an Evoloop.
        """
        super().__init__(rule_table={
            (0, 0, 0, 0, 1): 2,
            (1, 0, 2, 0, 2): 1,
            (1, 1, 2, 7, 2): 7,
            (2, 0, 1, 7, 2): 2,
            (2, 1, 3, 2, 2): 2,
            (4, 0, 1, 2, 5): 0,
            (0, 0, 0, 0, 4): 3,
            (1, 0, 2, 1, 1): 1,
            (1, 1, 2, 7, 3): 5,
            (2, 0, 2, 0, 2): 2,
            (2, 1, 4, 2, 2): 2,
            (4, 0, 1, 6, 2): 0,
            (0, 0, 0, 1, 2): 2,
            (1, 0, 2, 1, 2): 1,
            (1, 1, 3, 2, 2): 1,
            (2, 0, 2, 0, 3): 2,
            (2, 1, 6, 2, 2): 2,
            (4, 0, 2, 1, 2): 0,
            (0, 0, 0, 1, 5): 2,
            (1, 0, 2, 1, 3): 1,
            (1, 1, 3, 3, 2): 1,
            (2, 0, 2, 0, 5): 2,
            (2, 1, 7, 2, 2): 2,
            (4, 0, 2, 1, 5): 0,
            (0, 0, 0, 2, 1): 2,
            (1, 0, 2, 2, 1): 1,
            (1, 1, 5, 4, 2): 4,
            (2, 0, 2, 0, 6): 5,
            (2, 2, 2, 2, 4): 2,
            (4, 0, 2, 2, 2): 1,
            (0, 0, 0, 2, 4): 2,
            (1, 0, 2, 2, 4): 4,
            (1, 1, 5, 7, 2): 7,
            (2, 0, 2, 0, 7): 3,
            (2, 2, 2, 2, 7): 2,
            (4, 0, 2, 3, 2): 1,
            (0, 0, 0, 4, 2): 2,
            (1, 0, 2, 2, 7): 7,
            (1, 1, 6, 2, 4): 4,
            (2, 0, 2, 1, 2): 2,
            (2, 2, 2, 3, 4): 2,
            (4, 0, 2, 6, 2): 6,
            (0, 0, 0, 4, 5): 2,
            (1, 0, 2, 3, 2): 4,
            (1, 1, 6, 2, 7): 7,
            (2, 0, 2, 1, 5): 2,
            (2, 2, 2, 3, 7): 2,
            (4, 0, 3, 1, 2): 0,
            (0, 0, 0, 7, 5): 2,
            (1, 0, 2, 4, 1): 4,
            (1, 2, 2, 2, 4): 4,
            (2, 0, 2, 2, 1): 2,
            (2, 2, 2, 4, 3): 2,
            (4, 0, 3, 2, 2): 1,
            (0, 0, 1, 0, 2): 2,
            (1, 0, 2, 4, 2): 4,
            (1, 2, 2, 2, 7): 7,
            (2, 0, 2, 2, 2): 2,
            (2, 2, 2, 4, 4): 2,
            (5, 0, 0, 0, 2): 5,
            (0, 0, 2, 1, 4): 1,
            (1, 0, 2, 4, 3): 4,
            (1, 2, 2, 4, 3): 4,
            (2, 0, 2, 2, 3): 2,
            (2, 2, 2, 7, 3): 2,
            (5, 0, 0, 1, 2): 5,
            (0, 0, 2, 1, 7): 1,
            (1, 0, 2, 5, 1): 1,
            (1, 2, 2, 7, 3): 7,
            (2, 0, 2, 3, 2): 3,
            (2, 2, 2, 7, 7): 2,
            (5, 0, 0, 2, 1): 5,
            (0, 0, 2, 3, 2): 2,
            (1, 0, 2, 5, 2): 7,
            (1, 2, 3, 2, 4): 4,
            (2, 0, 2, 4, 2): 2,
            (2, 2, 3, 2, 4): 3,
            (5, 0, 0, 2, 3): 2,
            (0, 1, 1, 2, 2): 1,
            (1, 0, 2, 5, 4): 3,
            (1, 2, 3, 2, 7): 7,
            (2, 0, 2, 4, 5): 2,
            (2, 2, 3, 2, 7): 3,
            (5, 0, 0, 2, 4): 5,
            (0, 1, 2, 1, 2): 1,
            (1, 0, 2, 5, 7): 7,
            (1, 2, 4, 2, 6): 6,
            (2, 0, 2, 5, 2): 5,
            (3, 0, 0, 0, 1): 3,
            (5, 0, 0, 2, 7): 5,
            (0, 1, 2, 3, 2): 1,
            (1, 0, 2, 7, 1): 7,
            (1, 2, 4, 3, 3): 3,
            (2, 0, 2, 6, 2): 0,
            (3, 0, 0, 0, 2): 2,
            (5, 0, 0, 4, 2): 5,
            (0, 1, 2, 4, 2): 1,
            (1, 0, 2, 7, 2): 7,
            (1, 2, 6, 2, 7): 6,
            (2, 0, 2, 6, 5): 0,
            (3, 0, 0, 0, 3): 2,
            (5, 0, 0, 7, 2): 5,
            (0, 1, 2, 4, 5): 1,
            (1, 0, 2, 7, 3): 5,
            (2, 0, 0, 0, 1): 2,
            (2, 0, 2, 7, 2): 2,
            (3, 0, 0, 0, 4): 3,
            (5, 0, 2, 0, 2): 2,
            (0, 1, 2, 5, 2): 6,
            (1, 0, 5, 1, 2): 1,
            (2, 0, 0, 0, 2): 2,
            (2, 0, 2, 7, 5): 2,
            (3, 0, 0, 0, 7): 4,
            (5, 0, 2, 0, 5): 2,
            (0, 1, 2, 6, 2): 6,
            (1, 0, 5, 4, 2): 4,
            (2, 0, 0, 0, 4): 2,
            (2, 0, 3, 1, 2): 2,
            (3, 0, 0, 1, 2): 3,
            (5, 0, 2, 1, 2): 5,
            (0, 1, 2, 7, 2): 1,
            (1, 0, 5, 7, 2): 7,
            (2, 0, 0, 0, 5): 2,
            (2, 0, 3, 2, 2): 2,
            (3, 0, 0, 3, 2): 2,
            (5, 0, 2, 1, 5): 2,
            (0, 1, 2, 7, 5): 1,
            (1, 0, 6, 2, 1): 1,
            (2, 0, 0, 0, 6): 0,
            (2, 0, 3, 4, 2): 2,
            (3, 0, 0, 4, 2): 1,
            (5, 0, 2, 4, 2): 5,
            (0, 1, 3, 4, 2): 1,
            (1, 0, 6, 2, 4): 4,
            (2, 0, 0, 0, 7): 1,
            (2, 0, 3, 4, 5): 2,
            (3, 0, 1, 0, 2): 1,
            (5, 0, 2, 7, 2): 5,
            (0, 1, 3, 7, 2): 1,
            (1, 0, 6, 2, 7): 7,
            (2, 0, 0, 1, 2): 2,
            (2, 0, 3, 7, 2): 2,
            (3, 0, 1, 2, 5): 0,
            (5, 0, 3, 1, 2): 0,
            (0, 1, 4, 2, 2): 1,
            (1, 1, 1, 1, 2): 1,
            (2, 0, 0, 1, 5): 2,
            (2, 0, 4, 1, 2): 2,
            (3, 0, 2, 1, 2): 3,
            (6, 0, 2, 0, 2): 2,
            (0, 1, 4, 2, 5): 1,
            (1, 1, 1, 2, 2): 1,
            (2, 0, 0, 2, 1): 2,
            (2, 0, 4, 2, 2): 2,
            (3, 0, 2, 4, 2): 3,
            (6, 0, 2, 1, 2): 2,
            (0, 1, 4, 3, 2): 1,
            (1, 1, 1, 2, 4): 4,
            (2, 0, 0, 2, 2): 2,
            (2, 0, 4, 4, 2): 2,
            (3, 0, 2, 5, 2): 1,
            (6, 0, 2, 2, 2): 0,
            (0, 1, 4, 3, 5): 1,
            (1, 1, 1, 2, 5): 1,
            (2, 0, 0, 2, 3): 2,
            (2, 0, 5, 1, 2): 2,
            (3, 0, 2, 7, 2): 3,
            (6, 0, 2, 4, 2): 2,
            (0, 1, 4, 4, 2): 1,
            (1, 1, 1, 2, 7): 7,
            (2, 0, 0, 2, 4): 2,
            (2, 0, 5, 4, 2): 5,
            (3, 0, 3, 3, 2): 1,
            (6, 0, 2, 7, 2): 2,
            (0, 1, 4, 6, 2): 1,
            (1, 1, 1, 6, 2): 1,
            (2, 0, 0, 2, 6): 0,
            (2, 0, 5, 7, 2): 5,
            (3, 1, 2, 1, 2): 3,
            (6, 1, 2, 2, 2): 0,
            (0, 1, 7, 2, 2): 1,
            (1, 1, 2, 1, 2): 1,
            (2, 0, 0, 2, 7): 2,
            (2, 0, 6, 1, 2): 5,
            (3, 1, 2, 4, 2): 3,
            (6, 2, 2, 2, 4): 0,
            (0, 1, 7, 2, 5): 1,
            (1, 1, 2, 1, 3): 1,
            (2, 0, 0, 3, 2): 4,
            (2, 0, 6, 2, 1): 2,
            (3, 1, 2, 5, 2): 1,
            (6, 2, 2, 2, 7): 0,
            (0, 1, 7, 5, 6): 1,
            (1, 1, 2, 1, 5): 1,
            (2, 0, 0, 4, 2): 3,
            (2, 0, 6, 4, 2): 5,
            (3, 1, 2, 7, 2): 3,
            (7, 0, 1, 0, 2): 0,
            (0, 1, 7, 6, 2): 1,
            (1, 1, 2, 2, 2): 1,
            (2, 0, 0, 4, 5): 2,
            (2, 0, 6, 7, 2): 5,
            (3, 2, 4, 2, 4): 3,
            (7, 0, 1, 1, 2): 0,
            (0, 1, 7, 7, 2): 1,
            (1, 1, 2, 2, 4): 4,
            (2, 0, 0, 5, 4): 5,
            (2, 0, 7, 1, 2): 2,
            (3, 2, 4, 2, 5): 1,
            (7, 0, 1, 2, 2): 0,
            (1, 0, 0, 0, 1): 1,
            (1, 1, 2, 2, 7): 7,
            (2, 0, 0, 5, 7): 5,
            (2, 0, 7, 2, 2): 2,
            (3, 2, 4, 2, 7): 3,
            (7, 0, 1, 2, 5): 0,
            (1, 0, 0, 1, 2): 1,
            (1, 1, 2, 3, 2): 1,
            (2, 0, 0, 6, 2): 0,
            (2, 0, 7, 7, 2): 2,
            (3, 2, 5, 2, 7): 1,
            (7, 0, 1, 6, 2): 0,
            (1, 0, 0, 2, 1): 1,
            (1, 1, 2, 4, 2): 4,
            (2, 0, 0, 7, 2): 2,
            (2, 1, 1, 2, 2): 2,
            (3, 2, 7, 2, 7): 3,
            (7, 0, 2, 1, 2): 0,
            (1, 0, 0, 2, 4): 4,
            (1, 1, 2, 4, 3): 4,
            (2, 0, 0, 7, 5): 2,
            (2, 1, 2, 2, 2): 2,
            (4, 0, 0, 0, 0): 1,
            (7, 0, 2, 1, 5): 0,
            (1, 0, 0, 2, 7): 7,
            (1, 1, 2, 5, 2): 7,
            (2, 0, 1, 0, 2): 2,
            (2, 1, 2, 2, 3): 2,
            (4, 0, 0, 0, 2): 1,
            (7, 0, 2, 2, 2): 1,
            (1, 0, 1, 2, 1): 1,
            (1, 1, 2, 5, 4): 3,
            (2, 0, 1, 1, 2): 2,
            (2, 1, 2, 2, 4): 2,
            (4, 0, 1, 0, 2): 0,
            (7, 0, 2, 3, 2): 0,
            (1, 0, 1, 2, 4): 4,
            (1, 1, 2, 5, 7): 7,
            (2, 0, 1, 2, 2): 2,
            (2, 1, 2, 2, 7): 2,
            (4, 0, 1, 1, 2): 0,
            (7, 0, 2, 6, 2): 6,
            (1, 0, 1, 2, 7): 7,
            (1, 1, 2, 6, 2): 6,
            (2, 0, 1, 4, 2): 2,
            (2, 1, 2, 3, 2): 3,
            (4, 0, 1, 2, 2): 0,
            (7, 0, 3, 1, 2): 0,
        }, add_rotations=True)

    def __call__(self, n, c, t):
        """
        From:
        Sayama, H. (1998). Constructing evolutionary systems on a simple deterministic cellular automata space.
        PhD, University of Tokyo, Department of Information Science.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        current_activity = n[1][1]
        top = n[0][1]
        right = n[1][2]
        bottom = n[2][1]
        left = n[1][0]
        key = (current_activity, top, right, bottom, left)

        if key not in self._rule_table:
            trbl = (top, right, bottom, left)
            new_activity = None

            # Let 8->0 with no condition.
            if current_activity == 8:
                new_activity = 0

            # To all the undefined situations in whose four neighbourhood (TRBL) there is at least one site in state 8,
            # apply the following:
            if 8 in trbl:
                # Let 0,1->8 if there is at least one site in state 2,3,...,7 in its four neighbourhood (TRBL),
                # otherwise let 0->0 and 1->1
                if current_activity == 0 or current_activity == 1:
                    if np.any([i in trbl for i in (2, 3, 4, 5, 6, 7)]):
                        new_activity = 8
                    elif current_activity == 0:
                        new_activity = 0
                    elif current_activity == 1:
                        new_activity = 1

                # Let 2,3,5->0.
                if current_activity in (2, 3, 5):
                    new_activity = 0

                # Let 4,6,7->1.
                if current_activity in (4, 6, 7):
                    new_activity = 1

            # Clear up all the undefined situations by letting 0->0 and 1,2,...,7->8.
            if new_activity is None and current_activity == 0:
                new_activity = 0
            if new_activity is None and current_activity in (1, 2, 3, 4, 5, 6, 7):
                new_activity = 8

            return new_activity

        return self._rule_table[key]

    @staticmethod
    def init_species13_loop(dim, row, col):
        """
        Create the initial conditions by specifying the a loop of species 13 and its starting position (as given by the
        coordinates of the first cell of the first row of the loop).

        :param dim: a 2-tuple representing the dimensions (number of rows and columns) of the CA

        :param row: the row number of the loop

        :param col: the column number of the loop

        :return: the initial conditions
        """

        initial_conditions = np.zeros(dim, dtype=np.int32)

        # 1st row
        initial_conditions[row][col] = 2
        initial_conditions[row][col+1] = 2
        initial_conditions[row][col+2] = 2
        initial_conditions[row][col+3] = 2
        initial_conditions[row][col+4] = 2
        initial_conditions[row][col+5] = 2
        initial_conditions[row][col+6] = 2
        initial_conditions[row][col+7] = 2
        initial_conditions[row][col+8] = 2
        initial_conditions[row][col+9] = 2
        initial_conditions[row][col+10] = 2
        initial_conditions[row][col+11] = 2
        initial_conditions[row][col+12] = 2
        initial_conditions[row][col+13] = 2
        initial_conditions[row][col+14] = 2

        # 2nd row
        initial_conditions[row+1][col-1] = 2
        initial_conditions[row+1][col] = 0
        initial_conditions[row+1][col+1] = 1
        initial_conditions[row+1][col+2] = 7
        initial_conditions[row+1][col+3] = 0
        initial_conditions[row+1][col+4] = 1
        initial_conditions[row+1][col+5] = 7
        initial_conditions[row+1][col+6] = 0
        initial_conditions[row+1][col+7] = 1
        initial_conditions[row+1][col+8] = 7
        initial_conditions[row+1][col+9] = 0
        initial_conditions[row+1][col+10] = 1
        initial_conditions[row+1][col+11] = 4
        initial_conditions[row+1][col+12] = 0
        initial_conditions[row+1][col+13] = 1
        initial_conditions[row+1][col+14] = 4
        initial_conditions[row+1][col+15] = 2

        # 3rd row
        initial_conditions[row+2][col-1] = 2
        initial_conditions[row+2][col] = 7
        initial_conditions[row+2][col+1] = 2
        initial_conditions[row+2][col+2] = 2
        initial_conditions[row+2][col+3] = 2
        initial_conditions[row+2][col+4] = 2
        initial_conditions[row+2][col+5] = 2
        initial_conditions[row+2][col+6] = 2
        initial_conditions[row+2][col+7] = 2
        initial_conditions[row+2][col+8] = 2
        initial_conditions[row+2][col+9] = 2
        initial_conditions[row+2][col+10] = 2
        initial_conditions[row+2][col+11] = 2
        initial_conditions[row+2][col+12] = 2
        initial_conditions[row+2][col+13] = 2
        initial_conditions[row+2][col+14] = 0
        initial_conditions[row+2][col+15] = 2

        # 4th row
        initial_conditions[row+3][col-1] = 2
        initial_conditions[row+3][col] = 1
        initial_conditions[row+3][col+1] = 2
        initial_conditions[row+3][col+13] = 2
        initial_conditions[row+3][col+14] = 1
        initial_conditions[row+3][col+15] = 2

        # 5th row
        initial_conditions[row+4][col-1] = 2
        initial_conditions[row+4][col] = 0
        initial_conditions[row+4][col+1] = 2
        initial_conditions[row+4][col+13] = 2
        initial_conditions[row+4][col+14] = 1
        initial_conditions[row+4][col+15] = 2

        # 6th row
        initial_conditions[row+5][col-1] = 2
        initial_conditions[row+5][col] = 7
        initial_conditions[row+5][col+1] = 2
        initial_conditions[row+5][col+13] = 2
        initial_conditions[row+5][col+14] = 1
        initial_conditions[row+5][col+15] = 2

        # 7th row
        initial_conditions[row + 6][col - 1] = 2
        initial_conditions[row + 6][col] = 1
        initial_conditions[row + 6][col + 1] = 2
        initial_conditions[row + 6][col + 13] = 2
        initial_conditions[row + 6][col + 14] = 1
        initial_conditions[row + 6][col + 15] = 2

        # 8th row
        initial_conditions[row + 7][col - 1] = 2
        initial_conditions[row + 7][col] = 0
        initial_conditions[row + 7][col + 1] = 2
        initial_conditions[row + 7][col + 13] = 2
        initial_conditions[row + 7][col + 14] = 1
        initial_conditions[row + 7][col + 15] = 2

        # 9th row
        initial_conditions[row + 8][col - 1] = 2
        initial_conditions[row + 8][col] = 7
        initial_conditions[row + 8][col + 1] = 2
        initial_conditions[row + 8][col + 13] = 2
        initial_conditions[row + 8][col + 14] = 1
        initial_conditions[row + 8][col + 15] = 2

        # 10th row
        initial_conditions[row + 9][col - 1] = 2
        initial_conditions[row + 9][col] = 1
        initial_conditions[row + 9][col + 1] = 2
        initial_conditions[row + 9][col + 13] = 2
        initial_conditions[row + 9][col + 14] = 1
        initial_conditions[row + 9][col + 15] = 2

        # 11th row
        initial_conditions[row + 10][col - 1] = 2
        initial_conditions[row + 10][col] = 0
        initial_conditions[row + 10][col + 1] = 2
        initial_conditions[row + 10][col + 13] = 2
        initial_conditions[row + 10][col + 14] = 1
        initial_conditions[row + 10][col + 15] = 2

        # 12th row
        initial_conditions[row + 11][col - 1] = 2
        initial_conditions[row + 11][col] = 7
        initial_conditions[row + 11][col + 1] = 2
        initial_conditions[row + 11][col + 13] = 2
        initial_conditions[row + 11][col + 14] = 1
        initial_conditions[row + 11][col + 15] = 2

        # 13th row
        initial_conditions[row + 12][col - 1] = 2
        initial_conditions[row + 12][col] = 1
        initial_conditions[row + 12][col + 1] = 2
        initial_conditions[row + 12][col + 13] = 2
        initial_conditions[row + 12][col + 14] = 1
        initial_conditions[row + 12][col + 15] = 2

        # 14th row
        initial_conditions[row + 13][col - 1] = 2
        initial_conditions[row + 13][col] = 0
        initial_conditions[row + 13][col + 1] = 2
        initial_conditions[row + 13][col + 13] = 2
        initial_conditions[row + 13][col + 14] = 1
        initial_conditions[row + 13][col + 15] = 2

        # 15th row
        initial_conditions[row + 14][col - 1] = 2
        initial_conditions[row + 14][col] = 7
        initial_conditions[row + 14][col + 1] = 2
        initial_conditions[row + 14][col + 2] = 2
        initial_conditions[row + 14][col + 3] = 2
        initial_conditions[row + 14][col + 4] = 2
        initial_conditions[row + 14][col + 5] = 2
        initial_conditions[row + 14][col + 6] = 2
        initial_conditions[row + 14][col + 7] = 2
        initial_conditions[row + 14][col + 8] = 2
        initial_conditions[row + 14][col + 9] = 2
        initial_conditions[row + 14][col + 10] = 2
        initial_conditions[row + 14][col + 11] = 2
        initial_conditions[row + 14][col + 12] = 2
        initial_conditions[row + 14][col + 13] = 2
        initial_conditions[row + 14][col + 14] = 1
        initial_conditions[row + 14][col + 15] = 2
        initial_conditions[row + 14][col + 16] = 2
        initial_conditions[row + 14][col + 17] = 2
        initial_conditions[row + 14][col + 18] = 2
        initial_conditions[row + 14][col + 19] = 2
        initial_conditions[row + 14][col + 20] = 2
        initial_conditions[row + 14][col + 21] = 2
        initial_conditions[row + 14][col + 22] = 2
        initial_conditions[row + 14][col + 23] = 2
        initial_conditions[row + 14][col + 24] = 2
        initial_conditions[row + 14][col + 25] = 2
        initial_conditions[row + 14][col + 26] = 2
        initial_conditions[row + 14][col + 27] = 2
        initial_conditions[row + 14][col + 28] = 2

        # 16th row
        initial_conditions[row + 15][col - 1] = 2
        initial_conditions[row + 15][col] = 1
        initial_conditions[row + 15][col + 1] = 0
        initial_conditions[row + 15][col + 2] = 7
        initial_conditions[row + 15][col + 3] = 1
        initial_conditions[row + 15][col + 4] = 0
        initial_conditions[row + 15][col + 5] = 7
        initial_conditions[row + 15][col + 6] = 1
        initial_conditions[row + 15][col + 7] = 0
        initial_conditions[row + 15][col + 8] = 7
        initial_conditions[row + 15][col + 9] = 1
        initial_conditions[row + 15][col + 10] = 0
        initial_conditions[row + 15][col + 11] = 7
        initial_conditions[row + 15][col + 12] = 1
        initial_conditions[row + 15][col + 13] = 0
        initial_conditions[row + 15][col + 14] = 7
        initial_conditions[row + 15][col + 15] = 1
        initial_conditions[row + 15][col + 16] = 1
        initial_conditions[row + 15][col + 17] = 1
        initial_conditions[row + 15][col + 18] = 1
        initial_conditions[row + 15][col + 19] = 1
        initial_conditions[row + 15][col + 20] = 1
        initial_conditions[row + 15][col + 21] = 1
        initial_conditions[row + 15][col + 22] = 1
        initial_conditions[row + 15][col + 23] = 1
        initial_conditions[row + 15][col + 24] = 1
        initial_conditions[row + 15][col + 25] = 1
        initial_conditions[row + 15][col + 26] = 1
        initial_conditions[row + 15][col + 27] = 1
        initial_conditions[row + 15][col + 28] = 1
        initial_conditions[row + 15][col + 29] = 2

        # 17th row
        initial_conditions[row + 16][col] = 2
        initial_conditions[row + 16][col + 1] = 2
        initial_conditions[row + 16][col + 2] = 2
        initial_conditions[row + 16][col + 3] = 2
        initial_conditions[row + 16][col + 4] = 2
        initial_conditions[row + 16][col + 5] = 2
        initial_conditions[row + 16][col + 6] = 2
        initial_conditions[row + 16][col + 7] = 2
        initial_conditions[row + 16][col + 8] = 2
        initial_conditions[row + 16][col + 9] = 2
        initial_conditions[row + 16][col + 10] = 2
        initial_conditions[row + 16][col + 11] = 2
        initial_conditions[row + 16][col + 12] = 2
        initial_conditions[row + 16][col + 13] = 2
        initial_conditions[row + 16][col + 14] = 2
        initial_conditions[row + 16][col + 15] = 2
        initial_conditions[row + 16][col + 16] = 2
        initial_conditions[row + 16][col + 17] = 2
        initial_conditions[row + 16][col + 18] = 2
        initial_conditions[row + 16][col + 19] = 2
        initial_conditions[row + 16][col + 20] = 2
        initial_conditions[row + 16][col + 21] = 2
        initial_conditions[row + 16][col + 22] = 2
        initial_conditions[row + 16][col + 23] = 2
        initial_conditions[row + 16][col + 24] = 2
        initial_conditions[row + 16][col + 25] = 2
        initial_conditions[row + 16][col + 26] = 2
        initial_conditions[row + 16][col + 27] = 2
        initial_conditions[row + 16][col + 28] = 2

        return np.array([initial_conditions])



from .ca_functions import *


class HopfieldNet:
    """
    An implementation of the Hopfield network. Due to limitations of this implementation, only an odd number of cells
    is supported.

    For more information on Hopfield networks, see:

    .. code-block:: text

        Hopfield, J. J. (1982). Neural networks and physical systems with emergent collective
        computational abilities. Proceedings of the national academy of sciences, 79(8), 2554-2558.
    """
    def __init__(self, num_cells):
        """
        Create an instance of the Hopfield network.

        :param num_cells: the number of cells in this Hopfield network; only an odd number of cells are supported in
                          this implementation
        """
        self.apply_rule = AsynchronousRule(apply_rule=self._rule, num_cells=num_cells)
        self._r = num_cells // 2

    def train(self, P):
        """
        The training set consists of patterns to be learned by this net. The patterns should be composed of 
        bipolar ({-1,1}), and not binary ({0,1}), values.

        :param P: the set of training patterns
        """
        self._W = np.zeros((len(P[0]), len(P[0])), dtype=np.int32)
        for p in P:
            for i in range(len(p)):
                for j in range(len(p)):
                    if i ==j:
                        self._W[i, j] = 0
                    else:
                        self._W[i, j] += p[i]*p[j]

    def _rule(self, n, c, t):
        left_neighbours = n[0 : len(n)//2]
        right_neighbours = n[len(n)//2 + 1 :]
        V = 0
        for j, left_V in enumerate(left_neighbours):
            V += self._W[c - self._r + j, c] * left_V
        for j, right_V in enumerate(right_neighbours):
            V += self._W[(c + j + 1) % len(n), c] * right_V
        return 1 if V >= 0 else -1

    @property
    def W(self):
        """
        Returns the learned weight matrix.

        :return: the learned weight matrix
        """
        return self._W

    @property
    def r(self):
        """
        The radius of this automaton.

        :return: the radius
        """
        return self._r



from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib.collections as mcoll
import numpy as np


def plot2d(ca, timestep=None, title='', *, colormap='Greys', show_grid=False, show_margin=True, scale=0.6,
           show=True, **imshow_kwargs):
    """
    Plots the state of the given 2D cellular automaton at the given timestep. If no timestep is provided, then the last
    timestep is plotted.

    The `show_margin` argument controls whether or not a margin is displayed in the resulting plot. When `show_margin`
    is set to `False`, then the plot takes up the entirety of the window. The `scale` argument is only used when the
    `show_margins` argument is `False`. It controls the resulting scale (i.e. relative size) of the image when there
    are no margins.

    :param ca: the 2D cellular automaton to plot

    :param timestep: the timestep of interest

    :param title: the title to place on the plot

    :param colormap: the color map to use (default is "Greys")

    :param show_grid: whether to display a grid (default is False)

    :param show_margin: whether to display the margin (default is True)

    :param scale: the scale of the figure (default is 0.6)

    :param show: show the plot (default is True)

    :param imshow_kwargs: keyword arguments for the Matplotlib `imshow` function
    """
    cmap = plt.get_cmap(colormap)
    fig, ax = plt.subplots()
    plt.title(title)
    if not show_margin:
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
    if timestep is not None:
        data = ca[timestep]
    else:
        data = ca[-1]

    _add_grid_lines(ca, ax, show_grid)

    im = plt.imshow(data, interpolation='none', cmap=cmap, **imshow_kwargs)
    if not show_margin:
        baseheight, basewidth = im.get_size()
        fig.set_size_inches(basewidth*scale, baseheight*scale, forward=True)
    if show:
        plt.show()


def plot2d_slice(ca, slice=None, title='', *, colormap='Greys', show_grid=False, show_margin=True, scale=0.6,
                 show=True, **imshow_kwargs):
    """
    Plots a slice through the evolved states of a 2D cellular automaton. For example, consider the following `ca`,
    which may represent the evolution of a 3x3 2D cellular automaton over 3 timesteps:

    .. code-block:: text

        [[[ 1,  2,  3],
          [ 4,  5,  6],
          [ 7,  8,  9]],
         [[10, 11, 12],
          [13, 14, 15],
          [16, 17, 18]],
         [[19, 20, 21],
          [22, 23, 24],
          [25, 26, 27]]]

    By default, the following would be plotted:

    .. code-block:: text

        [[ 4,  5,  6],
         [13, 14, 15],
         [22, 23, 24]]

    If `0` is provided as the `slice` argument, then the following would be plotted:

    .. code-block:: text

        [[ 1,  2,  3],
         [10, 11, 12],
         [19, 20, 21]]

    The `show_margin` argument controls whether or not a margin is displayed in the resulting plot. When `show_margin`
    is set to `False`, then the plot takes up the entirety of the window. The `scale` argument is only used when the
    `show_margins` argument is `False`. It controls the resulting scale (i.e. relative size) of the image when there
    are no margins.

    :param ca: the 2D cellular automaton to plot

    :param slice: an int representing the index of the row to plot; by default, the "center" row is used

    :param title: the title to place on the plot

    :param colormap: the color map to use (default is "Greys")

    :param show_grid: whether to display a grid (default is False)

    :param show_margin: whether to display the margin (default is True)

    :param scale: the scale of the figure (default is 0.6)

    :param show: show the plot (default is True)

    :param imshow_kwargs: keyword arguments for the Matplotlib `imshow` function
    """
    cmap = plt.get_cmap(colormap)
    fig, ax = plt.subplots()
    plt.title(title)
    if not show_margin:
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
    if slice is not None:
        data = ca[:, slice]
    else:
        data = ca[:, len(ca[0])//2]

    _add_grid_lines(ca, ax, show_grid)

    im = plt.imshow(data, interpolation='none', cmap=cmap, **imshow_kwargs)
    if not show_margin:
        baseheight, basewidth = im.get_size()
        fig.set_size_inches(basewidth*scale, baseheight*scale, forward=True)
    if show:
        plt.show()


def plot2d_spacetime(ca, alpha=None, title='', show=True):
    """
    Plots the evolution of the given 2D cellular automaton as a 3D space-time plot.

    :param ca: the 2D cellular automaton to plot

    :param alpha: the alpha blending value; a real number between 0 (transparent) and 1 (opaque)

    :param title: the title to place on the plot

    :param show: show the plot (default is True)
    """
    fig = plt.figure(figsize=(10, 7))
    plt.title(title)
    ax = Axes3D(fig)
    ca = ca[::-1]
    xs = np.arange(ca.shape[2])[None, None, :]
    ys = np.arange(ca.shape[1])[None, :, None]
    zs = np.arange(ca.shape[0])[:, None, None]
    xs, ys, zs = np.broadcast_arrays(xs, ys, zs)
    masked_data = np.ma.masked_where(ca == 0, ca)
    ax.scatter(xs.ravel(),
               ys.ravel(),
               zs.ravel(),
               c=masked_data.ravel(), cmap='cool', marker='s', depthshade=False, alpha=alpha, edgecolors='#0F0F0F')
    if show:
        plt.show()


def plot2d_animate(ca, title='', *, colormap='Greys', show_grid=False, show_margin=True, scale=0.6, dpi=80,
                   interval=50, save=False, autoscale=False, show=True, **imshow_kwargs):
    """
    Animate the given 2D cellular automaton.

    The `show_margin` argument controls whether or not a margin is displayed in the resulting plot. When `show_margin`
    is set to `False`, then the plot takes up the entirety of the window. The `scale` argument is only used when the
    `show_margins` argument is `False`. It controls the resulting scale (i.e. relative size) of the image when there
    are no margins.

    The `dpi` argument represents the dots per inch of the animation when it is saved. There will be no visible effect
    of the `dpi` argument if the animation is not saved (i.e. when `save` is `False`).

    :param ca:  the 2D cellular automaton to animate

    :param title: the title to place on the plot (default is "")

    :param colormap: the color map to use (default is "Greys")

    :param show_grid: whether to display a grid (default is False)

    :param show_margin: whether to display the margin (default is True)

    :param scale: the scale of the figure (default is 0.6)

    :param dpi: the dots per inch of the image (default is 80)

    :param interval: the delay between frames in milliseconds (default is 50)

    :param save: whether to save the animation to a local file (default is False)

    :param autoscale: whether to autoscale the images in the animation; this should be set to True if the first
                      frame has a uniform value (e.g. all zeroes) (default is False)

    :param show: show the plot (default is True)

    :param imshow_kwargs: keyword arguments for the Matplotlib `imshow` function

    :return: the animation
    """
    cmap = plt.get_cmap(colormap)
    fig, ax = plt.subplots()
    plt.title(title)
    if not show_margin:
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)

    grid = _add_grid_lines(ca, ax, show_grid)

    im = plt.imshow(ca[0], animated=True, cmap=cmap, **imshow_kwargs)
    if not show_margin:
        baseheight, basewidth = im.get_size()
        fig.set_size_inches(basewidth*scale, baseheight*scale, forward=True)

    i = {'index': 0}
    def updatefig(*args):
        i['index'] += 1
        if i['index'] == len(ca):
            i['index'] = 0
        im.set_array(ca[i['index']])
        if autoscale:
            im.autoscale()
        return im, grid
    ani = animation.FuncAnimation(fig, updatefig, interval=interval, blit=True, save_count=len(ca))
    if save:
        ani.save('evolved.gif', dpi=dpi, writer="imagemagick")
    if show:
        plt.show()
    return ani


def _add_grid_lines(ca, ax, show_grid):
    """
    Adds grid lines to the plot.

    :param ca: the 2D cellular automaton to plot

    :param ax: the Matplotlib axis object

    :param show_grid: whether to display the grid lines

    :return: the grid object
    """
    grid_linewidth = 0.0
    if show_grid:
        plt.xticks(np.arange(-.5, len(ca[0][0]), 1), "")
        plt.yticks(np.arange(-.5, len(ca[0]), 1), "")
        plt.tick_params(axis='both', which='both', length=0)
        grid_linewidth = 0.5
    vertical = np.arange(-.5, len(ca[0][0]), 1)
    horizontal = np.arange(-.5, len(ca[0]), 1)
    lines = ([[(x, y) for y in (-.5, horizontal[-1])] for x in vertical] +
             [[(x, y) for x in (-.5, vertical[-1])] for y in horizontal])
    grid = mcoll.LineCollection(lines, linestyles='-', linewidths=grid_linewidth, color='grey')
    ax.add_collection(grid)

    return grid


def evolve2d(cellular_automaton, timesteps, apply_rule, r=1, neighbourhood='Moore', memoize=False):
    """
    Evolves the given cellular automaton for the specified time steps. Applies the given function to each cell during
    the evolution. A cellular automaton is represented here as an array of arrays, or matrix. This function expects
    an array containing the initial time step (i.e. initial condition, an array) for the cellular automaton. The final
    result is a matrix, where the number of rows equal the number of time steps specified.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step, or a history
                               of previous states, with the last item in the given array being the starting condition
                               for the evolution

    :param timesteps: the number of time steps in this evolution, or a callable that accepts the cellular automaton
                      (in terms of the history of its evolution) and the current timestep number, and is expected to
                      return a boolean indicating whether the evolution should continue; note that if a number is given,
                      this value refers to the total number of time steps in this cellular automaton evolution, which
                      includes the initial condition

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy
                       2D array of dimensions 2r+1 x 2r+1, representing the neighbourhood of the cell (if the
                       'von Neumann' neighbourhood is specified, the array will be a masked array); the cell identity,
                       which is a tuple representing the row and column indices of the cell in the cellular automaton
                       matrix, as (row, col); the time step, which is a scalar representing the time step in the
                       evolution

    :param r: the neighbourhood radius; the neighbourhood dimensions will be 2r+1 x 2r+1 (default is 1)

    :param neighbourhood: the neighbourhood type; valid values are 'Moore' or 'von Neumann' (default is 'Moore')

    :param memoize: allowed values are True, False, and "recursive"; if True, then the result of applying the rule on a
                    given neighbourhood will be cached, and used on subsequent invocations of the rule; if "recursive",
                    then a recursive memoized algorithm will be used, in which recursively wider neighbourhoods are
                    cached, along with the result of applying the rule on the cells in the widened neighbourhood; the
                    True and "recursive" options can result in a significant improvement to execution speed if the rule
                    is expensive to invoke; the "recursive" option works best when there are strongly repetitive
                    patterns in the CA, and when the state consists of 2^k x 2^k cells; if False, then no caching will
                    be used; NOTE: this should only be set to True or "recursive" for rules which do not store any state
                    upon invocation, and for rules which do not depend in the cell index or timestep number (default is
                    False)

    :return: a list of matrices, containing the results of the evolution, where the number of rows equal the number
             of time steps specified
    """
    von_neumann_mask = np.zeros((2*r + 1, 2*r + 1), dtype=bool)
    for i in range(len(von_neumann_mask)):
        mask_size = np.absolute(r - i)
        von_neumann_mask[i][:mask_size] = 1
        if mask_size != 0:
            von_neumann_mask[i][-mask_size:] = 1

    _, rows, cols = cellular_automaton.shape
    neighbourhood_indices = _get_neighbourhood_indices(rows, cols, r)

    cell_indices = _get_cell_indices((rows, cols))
    cell_idx_to_neigh_idx = _get_cell_indices_to_neighbourhood_indices(cell_indices, rows, cols, r)

    if callable(timesteps):
        return _evolve2d_dynamic(cellular_automaton, timesteps, apply_rule, neighbourhood, rows, cols,
                                 neighbourhood_indices, von_neumann_mask, memoize, cell_indices, cell_idx_to_neigh_idx)
    else:
        return _evolve2d_fixed(cellular_automaton, timesteps, apply_rule, neighbourhood, rows, cols,
                               neighbourhood_indices, von_neumann_mask, memoize, cell_indices, cell_idx_to_neigh_idx)


def _evolve2d_fixed(cellular_automaton, timesteps, apply_rule, neighbourhood, rows, cols,
                    neighbourhood_indices, von_neumann_mask, memoize, cell_indices, cell_idx_to_neigh_idx):
    """
    Evolves the given cellular automaton for a fixed of timesteps.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step, or a history
                               of previous states, with the last item in the given array being the starting condition
                               for the evolution

    :param timesteps: the number of time steps in this evolution; this value refers to the total number of time steps in
                      this cellular automaton evolution, which includes the initial condition

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy
                       2D array of dimensions 2r+1 x 2r+1, representing the neighbourhood of the cell (if the
                       'von Neumann' neighbourhood is specified, the array will be a masked array); the cell identity,
                       which is a tuple representing the row and column indices of the cell in the cellular automaton
                       matrix, as (row, col); the time step, which is a scalar representing the time step in the
                       evolution

    :param neighbourhood: the neighbourhood type; valid values are 'Moore' or 'von Neumann'

    :param rows: the number of rows in the CA

    :param cols: the number of columns in the CA

    :param neighbourhood_indices: the indices of cells, by neighbourhood

    :param von_neumann_mask: a numpy mask for von Neumann neighbourhoods

    :param memoize: whether to use memoization

    :param cell_indices: a 2D array containing 2-tuples representing the row and column coordinates of each cell

    :param cell_idx_to_neigh_idx: a dictionary mapping groups of cell indices to a group's neighbourhood indices

    :return: a list of matrices, containing the results of the evolution, where the number of rows equal the number
             of time steps specified
    """
    initial_conditions = cellular_automaton[-1]
    array = np.zeros((timesteps, rows, cols), dtype=cellular_automaton.dtype)
    array[0] = initial_conditions

    memo_table = {}
    recursive_cache = _MemoizationCache()

    for t in range(1, timesteps):
        cell_layer = array[t - 1]

        if memoize is "recursive":
            next_state = np.zeros(cell_layer.shape, dtype=cell_layer.dtype)
            _step(cell_indices, cell_idx_to_neigh_idx, cell_layer, next_state, recursive_cache, apply_rule,
                  neighbourhood, von_neumann_mask, t)
            array[t] = next_state
        else:
            for row, cell_row in enumerate(cell_layer):
                for col, cell in enumerate(cell_row):
                    n = _get_neighbourhood(cell_layer, neighbourhood_indices, row, col, neighbourhood, von_neumann_mask)
                    if memoize is True:
                        array[t][row][col] = _get_memoized(n, (row, col), t, apply_rule, memo_table)
                    elif memoize is False:
                        array[t][row][col] = apply_rule(n, (row, col), t)
                    else:
                        raise Exception("unsupported memoization option: %s" % memoize)

    return np.concatenate((cellular_automaton, array[1:]), axis=0)


def _evolve2d_dynamic(cellular_automaton, timesteps, apply_rule, neighbourhood, rows, cols,
                    neighbourhood_indices, von_neumann_mask, memoize, cell_indices, cell_idx_to_neigh_idx):
    """
    Evolves the given cellular automaton for a dynamic number of timesteps.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step, or a history
                               of previous states, with the last item in the given array being the starting condition
                               for the evolution

    :param timesteps: a callable that accepts the cellular automaton (in terms of the history of its evolution) and the
                      current timestep number, and is expected to return a boolean indicating whether the evolution
                      should continue

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy
                       2D array of dimensions 2r+1 x 2r+1, representing the neighbourhood of the cell (if the
                       'von Neumann' neighbourhood is specified, the array will be a masked array); the cell identity,
                       which is a tuple representing the row and column indices of the cell in the cellular automaton
                       matrix, as (row, col); the time step, which is a scalar representing the time step in the
                       evolution

    :param neighbourhood: the neighbourhood type; valid values are 'Moore' or 'von Neumann'

    :param rows: the number of rows in the CA

    :param cols: the number of columns in the CA

    :param neighbourhood_indices: the indices of cells, by neighbourhood

    :param von_neumann_mask: a numpy mask for von Neumann neighbourhoods

    :param memoize: whether to use memoization

    :param cell_indices: a 2D array containing 2-tuples representing the row and column coordinates of each cell

    :param cell_idx_to_neigh_idx: a dictionary mapping groups of cell indices to a group's neighbourhood indices

    :return: a list of matrices, containing the results of the evolution, where the number of rows equal the number
             of time steps specified
    """
    initial_conditions = cellular_automaton[-1]
    array = [initial_conditions]

    memo_table = {}
    recursive_cache = _MemoizationCache()

    t = 1
    while timesteps(np.array(array), t):
        prev_layer = array[-1]
        next_layer = np.zeros((rows, cols), dtype=cellular_automaton.dtype)

        if memoize is "recursive":
            _step(cell_indices, cell_idx_to_neigh_idx, prev_layer, next_layer, recursive_cache, apply_rule,
                  neighbourhood, von_neumann_mask, t)
        else:
            for row, cell_row in enumerate(prev_layer):
                for col, cell in enumerate(cell_row):
                    n = _get_neighbourhood(prev_layer, neighbourhood_indices, row, col, neighbourhood, von_neumann_mask)
                    if memoize is True:
                        next_layer[row][col] = _get_memoized(n, (row, col), t, apply_rule, memo_table)
                    elif memoize is False:
                        next_layer[row][col] = apply_rule(n, (row, col), t)
                    else:
                        raise Exception("unsupported memoization option: %s" % memoize)

        array.append(next_layer)
        t += 1

    return np.concatenate((cellular_automaton, array[1:]), axis=0)


def _get_neighbourhood_indices(rows, cols, r):
    """
    Returns a dictionary mapping the coordinates of a cell in a 2D CA to its neighbourhood indices.

    :param rows: the number of rows in the 2D CA

    :param cols: the number of columns in the 2D CA

    :param r: the radius of the neighbourhood

    :return: a dictionary, where the key is a 2-tuple, (row, col),
             and the value is a 2-tuple, (row_indices, col_indices)
    """
    indices = {}
    for row in range(rows):
        for col in range(cols):
            row_indices = range(row - r, row + r + 1)
            row_indices = [i - rows if i > (rows - 1) else i for i in row_indices]
            col_indices = range(col - r, col + r + 1)
            col_indices = [i - cols if i > (cols - 1) else i for i in col_indices]
            indices[(row, col)] = (row_indices, col_indices)
    return indices


def _get_neighbourhood(cell_layer, neighbourhood_indices, row, col, neighbourhood, von_neumann_mask):
    """
    Returns the cell neighbourhood for the cell given by the row and column index. If the neighbourhood is
    `von Neumann`, then an appropriately masked array is returned.

    :param cell_layer: an array with dimensions 2r+1 x 2r+1

    :param neighbourhood_indices: a 2-tuple containing the row and column indices of the neighbours of the cell given
                                  by the row and column index

    :param row: the row index of the cell

    :param col: the column index of the cell

    :param neighbourhood: the neighbourhood type

    :param von_neumann_mask: a boolean array with dimensions 2r+1 x 2r+1 representing which cells in the neighbourhood
                             should be masked

    :return: a 2r+1 x 2r+1 array representing the cell neighbourhood of the cell given by row and col, if the
             neighbourhood type is `von Neumann`, then the array will be masked
    """
    row_indices, col_indices = neighbourhood_indices[(row, col)]
    n = cell_layer[np.ix_(row_indices, col_indices)]
    if neighbourhood == 'Moore':
        return n
    elif neighbourhood == 'von Neumann':
        return np.ma.masked_array(n, von_neumann_mask)
    else:
        raise ValueError('unknown neighbourhood type: %s' % neighbourhood)


def _get_memoized(n, c, t, apply_rule, memoization_table):
    """
    Checks if the result of `apply_rule` is in the memoization table according to the neighbourhood, `n`,
    and returns the associated value if it is. Otherwise, `apply_rule` is invoked and the result is stored in
    the memoization table, and then returned.

    :param n: a numpy array representing the neighbourhood

    :param c: an int representing the current cell's index

    :param t: an int, representing the current timestep

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy
                       2D array of dimensions 2r+1 x 2r+1, representing the neighbourhood of the cell (if the
                       'von Neumann' neighbourhood is specified, the array will be a masked array); the cell identity,
                       which is a tuple representing the row and column indices of the cell in the cellular automaton
                       matrix, as (row, col); the time step, which is a scalar representing the time step in the
                       evolution

    :param memoization_table: a dictionary mapping a neighbourhood to the rule result for such a neighbourhood

    :return: the result of invoking the given rule with the given n, c, and t arguments
    """
    key = n.tobytes()
    if key in memoization_table:
        return memoization_table[key]
    else:
        result = apply_rule(n, c, t)
        memoization_table[key] = result
        return result


def _get_cell_indices(shape):
    """
    Returns a 2D array containing 2-tuples representing the row and column coordinates of each cell.

    :param shape: a 2-tuple of ints representing the number of rows and columns

    :return: a 2D array containing 2-tuples representing the row and column coordinates of each cell
    """
    indices = []
    rows, cols = shape[0], shape[1]
    for r in range(rows):
        row = []
        for c in range(cols):
            row.append((r, c))
        indices.append(row)
    return np.array(indices, dtype="i, i")


def _get_cell_indices_to_neighbourhood_indices(cell_indices, n_rows, n_cols, r):
    """
    Returns a dictionary which maps groups of cell indices (i.e. 2-tuples of ints representing the row and colummn
    coordinates of cells) to a group's neighbourhood indices. The groups of cells are obtained by recursively dividing
    a 2D array of cell indices into quadrants.

    :param cell_indices: a 2D array containing 2-tuples representing the row and column coordinates of each cell

    :param n_rows: an int representing the number of rows in the 2D CA

    :param n_cols: an int representing the number of columns in the 2D CA

    :param r: an int representing the neighbourhood radius of the 2D CA

    :return: a dictionary mapping groups of cell indices to a group's neighbourhood indices
    """
    cell_indices_to_neighbourhood_indices = {}

    if cell_indices.size == 0:
        return cell_indices_to_neighbourhood_indices

    row_start = cell_indices[0][0][0]
    row_end = cell_indices[-1][0][0]
    col_start = cell_indices[0][0][1]
    col_end = cell_indices[0][-1][1]
    row_indices = range(row_start - r, row_end + r + 1)
    row_indices = [i - n_rows if i > (n_rows - 1) else i for i in row_indices]
    col_indices = range(col_start - r, col_end + r + 1)
    col_indices = [i - n_cols if i > (n_cols - 1) else i for i in col_indices]

    cell_indices_to_neighbourhood_indices[cell_indices.tobytes()] = (row_indices, col_indices)

    if cell_indices.shape[0] > 1 or cell_indices.shape[1] > 1:
        nw_indices, sw_indices, ne_indices, se_indices = _get_sub_matrices(cell_indices)
        cell_indices_to_neighbourhood_indices.update(_get_cell_indices_to_neighbourhood_indices(nw_indices, n_rows, n_cols, r))
        cell_indices_to_neighbourhood_indices.update(_get_cell_indices_to_neighbourhood_indices(ne_indices, n_rows, n_cols, r))
        cell_indices_to_neighbourhood_indices.update(_get_cell_indices_to_neighbourhood_indices(sw_indices, n_rows, n_cols, r))
        cell_indices_to_neighbourhood_indices.update(_get_cell_indices_to_neighbourhood_indices(se_indices, n_rows, n_cols, r))

    return cell_indices_to_neighbourhood_indices


def _get_sub_matrices(arr):
    """
    Takes the given 2D array of 2-tuples representing cell coordinates, and partitions it as evenly as possible into
    4 quadrants, returning the northwest, southwest, northeast, and southeast quadrants as separate 2D arrays, in that
    order.

    :param arr: a 2D array of 2-tuples representing cell coordinates

    :return: the quadrants of the given array as separate 2D arrays, in the order northwest, southwest, northeast,
             and southeast
    """
    indices = []
    indices.extend([x.tolist() for x in np.array_split(np.array_split(arr, 2, axis=1)[0], 2, axis=0)])
    indices.extend([x.tolist() for x in np.array_split(np.array_split(arr, 2, axis=1)[1], 2, axis=0)])
    return np.array(indices[0], dtype="i, i"), np.array(indices[1], dtype="i, i"), \
           np.array(indices[2], dtype="i, i"), np.array(indices[3], dtype="i, i")


def _update_state(cell_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t):
    """
    Perform an update on the given next state using the current state and memoization cache.

    :param cell_indices: a 2D array containing 2-tuples representing the row and column coordinates of each cell

    :param cell_idx_to_neigh_idx: a dictionary mapping groups of cell indices to a group's neighbourhood indices

    :param curr_state: the current state (i.e. the state after the previous timestep)

    :param next_state: the next state (i.e. the result after the current timestep)

    :param cache: a mapping from state neighbourhoods to their activities

    :param apply_rule: the rule to apply during each cell update

    :param neighbourhood_type: the neighbourhood type; valid values are 'Moore' or 'von Neumann'

    :param von_neumann_mask: :param von_neumann_mask: a numpy mask for von Neumann neighbourhoods

    :param t: the current timestep
    """
    if cell_indices.size == 0:
        return

    neigh_row_indices, neigh_col_indices = cell_idx_to_neigh_idx[cell_indices.tobytes()]
    state = curr_state[np.ix_(neigh_row_indices, neigh_col_indices)]

    if state in cache:
        # update next_state with next vals from cache
        state_row_indices = neigh_row_indices[1:-1]
        state_col_indices = neigh_col_indices[1:-1]
        next_state[np.ix_(state_row_indices, state_col_indices)] = cache[state]
    else:
        if cell_indices.shape[0] > 1 or cell_indices.shape[1] > 1:
            _step(cell_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t)
        else:
            # invoke rule and update next_state for cell
            c = cell_indices[0][0]
            neighbourhood = curr_state[np.ix_(neigh_row_indices, neigh_col_indices)]

            if neighbourhood_type == 'von Neumann':
                neighbourhood = np.ma.masked_array(neighbourhood, von_neumann_mask)
            elif neighbourhood_type != 'Moore':
                raise ValueError('unknown neighbourhood type: %s' % neighbourhood_type)

            val = apply_rule(neighbourhood, c, t)
            next_state[c[0]][c[1]] = val
        # get the result from the next_state for the left_indices and place in cache
        state_row_indices = neigh_row_indices[1:-1]
        state_col_indices = neigh_col_indices[1:-1]
        vals_to_cache = next_state[np.ix_(state_row_indices, state_col_indices)]
        cache[state] = vals_to_cache


def _step(cell_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t):
    """
    Perform an update on the given next state using the current state and memoization cache, based on
    an even 4-way split of the cell indices.

    :param cell_indices: a 2D array containing 2-tuples representing the row and column coordinates of each cell

    :param cell_idx_to_neigh_idx: a dictionary mapping groups of cell indices to a group's neighbourhood indices

    :param curr_state: the current state (i.e. the state after the previous timestep)

    :param next_state: the next state (i.e. the result after the current timestep)

    :param cache: a mapping from state neighbourhoods to their activities

    :param apply_rule: the rule to apply during each cell update

    :param neighbourhood_type: the neighbourhood type; valid values are 'Moore' or 'von Neumann'

    :param von_neumann_mask: :param von_neumann_mask: a numpy mask for von Neumann neighbourhoods

    :param t: the current timestep
    """
    nw_indices, sw_indices, ne_indices, se_indices = _get_sub_matrices(cell_indices)
    _update_state(nw_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t)
    _update_state(ne_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t)
    _update_state(sw_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t)
    _update_state(se_indices, cell_idx_to_neigh_idx, curr_state, next_state, cache, apply_rule, neighbourhood_type,
                  von_neumann_mask, t)


def init_simple2d(rows, cols, val=1, dtype=np.int32, coords=None):
    """
    Returns a matrix initialized with zeroes, with its center value set to the specified value, or 1 by default.
    If the `coords` argument is specified, then the specified cell at the given coordinates will have its value
    set to `val`, otherwise the center cell will be set.

    :param rows: the number of rows in the matrix

    :param cols: the number of columns in the matrix

    :param val: the value to be used in the center of the matrix (1, by default)

    :param dtype: the data type (np.int32 by default)

    :param coords: a 2-tuple specifying the row and column of the cell to be initialized (None by default)

    :return: a tensor with shape (1, rows, cols), with the center value initialized to the specified value, or 1 by default
    """
    x = np.zeros((rows, cols), dtype=dtype)
    if coords is not None:
        if not isinstance(coords, (tuple, list)) or len(coords) != 2:
            raise TypeError("coords must be a list or tuple of length 2")
        x[coords[0]][coords[1]] = val
    else:
        x[x.shape[0]//2][x.shape[1]//2] = val
    return np.array([x])


def init_random2d(rows, cols, k=2, dtype=np.int32):
    """
    Returns a randomly initialized matrix with values consisting of numbers in {0,...,k - 1}, where k = 2 by default.
    If dtype is not an integer type, then values will be uniformly distributed over the half-open interval [0, k - 1).

    :param rows: the number of rows in the matrix

    :param cols: the number of columns in the matrix

    :param k: the number of states in the cellular automaton (2, by default)

    :param dtype: the data type

    :return: a tensor with shape (1, rows, cols), randomly initialized with numbers in {0,...,k - 1}
    """
    if np.issubdtype(dtype, np.integer):
        rand_nums = np.random.randint(k, size=(rows, cols), dtype=dtype)
    else:
        rand_nums = np.random.uniform(0, k - 1, size=(rows, cols)).astype(dtype)
    return np.array([rand_nums])


def game_of_life_rule(neighbourhood, c, t):
    """
    Conway's Game of Life rule.

    :param neighbourhood: the current cell's neighbourhood

    :param c: the index of the current cell

    :param t: the current timestep

    :return: the state of the current cell at the next timestep
    """
    center_cell = neighbourhood[1][1]
    total = np.sum(neighbourhood)
    if center_cell == 1:
        if total - 1 < 2:
            return 0  # Any live cell with fewer than two live neighbours dies, as if caused by underpopulation.
        if total - 1 == 2 or total - 1 == 3:
            return 1  # Any live cell with two or three live neighbours lives on to the next generation.
        if total - 1 > 3:
            return 0  # Any live cell with more than three live neighbours dies, as if by overpopulation.
    else:
        if total == 3:
            return 1  # Any dead cell with exactly three live neighbours becomes a live cell, as if by reproduction.
        else:
            return 0


class _MemoizationCache(object):
    """
    A cache that maps 2D numpy arrays to values. The challenge is to inexpensively compute a unique hash for a 2D
    NumPy array, so that it can be used to quickly look up corresponding values. The NumPy array function `tobytes()`
    provides a fast method of producing a unique key for each array, however, it is indifferent to array shape. Thus,
    two arrays with the same contents, but different shapes, would produce the same key. Here, we use an underlying
    dictionary to keep track of collisions (if any) that may occur when two arrays with the same contents but different
    shapes are added to the cache. If a collision occurs, then all values are kept and associated with the originating
    array's shape. At read time, the values for a given array/key are disambiguated using the array's shape.
    """
    def __init__(self):
        self.hashmap = {}

    def put(self, array, val):
        hash = array.tobytes()
        shape = array.shape
        if hash not in self.hashmap:
            self.hashmap[hash] = []
        self.hashmap[hash].append((shape, val))

    def get(self, array):
        hash = array.tobytes()
        shape = array.shape
        collisions = self.hashmap[hash]
        if len(collisions) == 1:
            return collisions[0][1]
        else:
            for collision in collisions:
                collision_shape = collision[0]
                if collision_shape == shape:
                    return collision[1]
            raise Exception("could not find an entry with shape: %s" % shape)

    def __contains__(self, item):
        hash = item.tobytes()
        if hash not in self.hashmap:
            return False
        for collision in self.hashmap[hash]:
            if collision[0] == item.shape:
                return True
        return False

    def __iter__(self):
        for hash in self.hashmap:
            yield hash

    def __len__(self):
        return len(self.hashmap)

    def __setitem__(self, key, value):
        self.put(key, value)

    def __getitem__(self, key):
        return self.get(key)



import numpy as np
from .langtons_loop import LangtonsLoop


class SDSRLoop(LangtonsLoop):
    """
    An implementation of H. Sayama's SDSR loop. For more information, see:

    .. code-block:: text

       Sayama, H. (1998). Constructing evolutionary systems on a simple deterministic cellular automata space.
       PhD, University of Tokyo, Department of Information Science.
    """
    def __init__(self):
        """
        Create an SDSR Loop.
        """
        super().__init__()

        # Define rule '11152->8' and its rotationally symmetric ones.
        self._rule_table[(1, 1, 1, 5, 2)] = 8
        self._rule_table[(1, 2, 1, 1, 5)] = 8
        self._rule_table[(1, 5, 2, 1, 1)] = 8
        self._rule_table[(1, 1, 5, 2, 1)] = 8

    def __call__(self, n, c, t):
        """
        From:
        Sayama, H. (1998). Constructing evolutionary systems on a simple deterministic cellular automata space.
        PhD, University of Tokyo, Department of Information Science.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        current_activity = n[1][1]
        top = n[0][1]
        right = n[1][2]
        bottom = n[2][1]
        left = n[1][0]
        key = (current_activity, top, right, bottom, left)

        if key not in self._rule_table:
            trbl = (top, right, bottom, left)
            in_tube = self._is_in_tube(top, right, bottom, left)
            new_activity = None

            # Let 0->1 if it is in the tube and next to 1. Let all other 0s remain as is.
            if current_activity == 0:
                if in_tube and 1 in trbl:
                    new_activity = 1
                else:
                    new_activity = 0

            # Let 1->7 if it is in the tube and next to 7. Else, let 1->6 if it is in the tube and next to 6.
            # Else, let 1->4 if it is in the tube and next to 4.
            if current_activity == 1 and in_tube:
                if 7 in trbl:
                    new_activity = 7
                elif 6 in trbl:
                    new_activity = 6
                elif 4 in trbl:
                    new_activity = 4

            # Let 4,6,7->0 if it is in the tube and next to 0.
            if current_activity in (4, 6, 7) and in_tube and 0 in trbl:
                new_activity = 0

            # Let 2->1 if it is next to 3. Else, let 2 remain as is if it is next to another 2.
            if current_activity == 2 and 3 in trbl:
                new_activity = 1
            elif current_activity == 2 and 2 in trbl:
                new_activity = 2

            # Let 8->0 with no condition.
            if current_activity == 8:
                new_activity = 0

            # To all the undefined situations in whose four neighbourhood (TRBL) there is at least one site in state 8,
            # apply the following:
            if 8 in trbl:
                # Let 0,1->8 if there is at least one site in state 2,3,...,7 in its four neighbourhood (TRBL),
                # otherwise let 0->0 and 1->1
                if current_activity == 0 or current_activity == 1:
                    if np.any([i in trbl for i in (2, 3, 4, 5, 6, 7)]):
                        new_activity = 8
                    elif current_activity == 0:
                        new_activity = 0
                    elif current_activity == 1:
                        new_activity = 1

                # Let 2,3,5->0.
                if current_activity in (2, 3, 5):
                    new_activity = 0

                # Let 4,6,7->1.
                if current_activity in (4, 6, 7):
                    new_activity = 1

            # Clear up all the undefined situations by letting 0->0 and 1,2,...,7->8.
            if new_activity is None and current_activity == 0:
                new_activity = 0
            if new_activity is None and current_activity in (1, 2, 3, 4, 5, 6, 7):
                new_activity = 8

            return new_activity

        return self._rule_table[key]

    def _is_in_tube(self, top, right, bottom, left):
        k = 0
        for site in [top, right, bottom, left]:
            if site in (1, 2, 4, 6, 7):
                k += 1
        return k >= 2



import random

import numpy as np


def table_rule(neighbourhood, table):
    """
    A rule where the state is converted into a string, and looked up in the given table, to yield the return value.

    :param neighbourhood: a k-color array of length 2r + 1

    :param table: a table (map) of string representations of each neighbourhood state to the associated next
           cell state value; for example, for k = 2 and r = 2, a valid table might be: 
           {'101': 1, '111': 0, '011': 0, '110': 1, '000': 0, '100': 0, '010': 0, '001': 1}

    :return: a number, from 0 to k - 1, associated with the given state as specified in the given table
    """
    state_repr = ''.join(str(x) for x in neighbourhood)
    if not state_repr in table:
        raise ValueError("could not find state '%s' in table" % state_repr)
    return table[state_repr]


def random_rule_table(k, r, lambda_val=None, quiescent_state=None, strong_quiescence=False, isotropic=False):
    """
    Constructs and returns a random rule table using the "random-table" method, as described in:

    .. code-block:: text

        Langton, C. G. (1990). Computation at the edge of chaos: phase transitions
        and emergent computation.
        Physica D: Nonlinear Phenomena, 42(1-3), 12-37.

    :param k: the number of cell states

    :param r: the radius of the cellular automaton neighbourhood

    :param lambda_val: a real number in (0., 1.), representing the value of lambda; if None, a default value of
                       1.0 - 1/k will be used, where all states will be represented equally in the rule table

    :param quiescent_state: the state, a number in {0,...,k - 1}, to use as the quiescent state

    :param strong_quiescence: if True, all neighbourhood states uniform in cell state i will map to cell state i

    :param isotropic: if True, all planar rotations of a neighbourhood state will map to the same cell state

    :return: a tuple containing: a table describing a rule, constructed using the "random-table" table method as
             described by C. G. Langton, the actual lambda value, and the quiescent state used
    """
    states = []
    n = 2*r + 1
    for i in range(0, k**n):
        states.append(np.base_repr(i, k).zfill(n))
    table = {}
    if lambda_val is None:
        lambda_val = 1. - (1. / k)
    if quiescent_state is None:
        quiescent_state = np.random.randint(k, dtype=np.int32)
    if not (0 <= quiescent_state <= k - 1):
        raise ValueError("quiescent state must be a number in {0,...,k - 1}")
    other_states = [x for x in range(0, k) if x != quiescent_state]
    quiescent_state_count = 0
    for state in states:
        if strong_quiescence and len(set(state)) == 1:
            # if the cell states in neighbourhood are all the same, e.g. '111'
            cell_state = int(state[0], k)
            if cell_state == quiescent_state: quiescent_state_count += 1
        else:
            state_reversed = state[::-1]
            if isotropic and state_reversed in table:
                cell_state = table[state_reversed]
                if cell_state == quiescent_state: quiescent_state_count += 1
            else:
                if random.random() < (1. - lambda_val):
                    cell_state = quiescent_state
                    quiescent_state_count += 1
                else:
                    cell_state = random.choice(other_states)
        table[state] = cell_state
    actual_lambda_val = (k**n - quiescent_state_count) / k**n
    return table, actual_lambda_val, quiescent_state


def table_walk_through(rule_table, lambda_val, k, r, quiescent_state, strong_quiescence=False, isotropic=False):
    """
    Perturbs the given rule table using the "table-walk-through" approach described in:

    .. code-block:: text

        Langton, C. G. (1990). Computation at the edge of chaos: phase transitions
        and emergent computation.
        Physica D: Nonlinear Phenomena, 42(1-3), 12-37.

    The table's actual lambda value will be increased or decreased, incrementally and stochastically, until it reaches
    the given lambda value.

    :param rule_table: a table (map) of string representations of each neighbourhood state to the associated next 
                       cell state value; for example, for k = 2 and r = 2, a valid table might be: 
                       {'101': 1, '111': 0, '011': 0, '110': 1, '000': 0, '100': 0, '010': 0, '001': 1}

    :param lambda_val: a real number in (0., 1.), representing the value of lambda

    :param k: the number of cell states

    :param r: the radius of the cellular automaton neighbourhood

    :param quiescent_state: the state, a number in {0,...,k - 1}, to use as the quiescent state

    :param strong_quiescence: if True, all neighbourhood states uniform in cell state i will map to cell state i

    :param isotropic: if True, all planar rotations of a neighbourhood state will map to the same cell state

    :return: a tuple containing: a table describing a rule, constructed using the "table-walk-through" method as
             described by C. G. Langton, the actual lambda value
    """
    def actual_lambda():
        n = 2*r + 1
        transitions_to_quiescent_state = list(rule_table.values()).count(quiescent_state)
        return (k**n - transitions_to_quiescent_state) / k**n
    actual_lambda_val = actual_lambda()
    if actual_lambda_val == lambda_val:
        return rule_table, actual_lambda_val
    if actual_lambda_val > lambda_val:
        # reduce lambda
        attempts = 0
        while actual_lambda() > lambda_val and attempts < len(rule_table):
            attempts += 1
            states_to_others = [k for k in rule_table.keys() if rule_table[k] != quiescent_state]
            if strong_quiescence:
                # remove states that are all the same (i.e. '111'); presumably the strong quiescence condition is already
                #  enforced in the incoming rule table, and so these states shouldn't be changed
                states_to_others = [s for s in states_to_others if len(set(s)) != 1]
            if len(states_to_others) == 0:
                break
            state_to_perturb = random.choice(states_to_others)
            rule_table[state_to_perturb] = quiescent_state
            if isotropic:
                state_to_perturb_reversed = state_to_perturb[::-1]
                rule_table[state_to_perturb_reversed] = rule_table[state_to_perturb]
    elif actual_lambda_val < lambda_val:
        # increase lambda
        attempts = 0
        while actual_lambda() < lambda_val and attempts < len(rule_table):
            attempts += 1
            states_to_quiescent = [k for k in rule_table.keys() if rule_table[k] == quiescent_state]
            if strong_quiescence:
                # remove states that are all the same (i.e. '111')
                states_to_quiescent = [s for s in states_to_quiescent if len(set(s)) != 1]
            if len(states_to_quiescent) == 0:
                break
            state_to_perturb = random.choice(states_to_quiescent)
            other_states = [x for x in range(0, k) if x != quiescent_state]
            rule_table[state_to_perturb] = random.choice(other_states)
            if isotropic:
                state_to_perturb_reversed = state_to_perturb[::-1]
                rule_table[state_to_perturb_reversed] = rule_table[state_to_perturb]
    return rule_table, actual_lambda()



from .ca_functions import BaseRule


class Sandpile(BaseRule):
    """
    A rule that operates on von Neumann neighbourhoods with a radius of 1. A sandpile is a 2D cellular automaton and
    dynamical system that displays self-organized criticality. It was introduced by Bak, Tang and Wiesenfeld in 1987.
    """
    def __init__(self, rows, cols, is_closed_boundary=True):
        """
        Creates a Sandpile.

        :param rows: the number of rows in this 2D CA

        :param cols: the number of columns in this 2D CA

        :param is_closed_boundary: whether or not the sandpile's boundary is closed; if it is closed, then all the
               boundary cells will maintain a value of 0 (default is True)
        """
        self._K = 4  # this value is hard-coded because the neighbourhood type, "von Neumann", is fixed
        self._rows = rows
        self._cols = cols
        self._is_closed_boundary = is_closed_boundary
        self._grain_additions = []

    def add_grain(self, cell_index, timestep):
        """
        Drop a grain of sand at the given cell in the given timestep.

        :param cell_index: a 2-tuple representing the row index and column index of the cell that will have
                           a grain of sand added

        :param timestep: the timestep at which the grain addition will occur
        """
        self._grain_additions.append(_GrainAddition(cell_index, timestep))

    def _is_in_boundary(self, c):
        """
        Returns True if the given cell (as a 2-tuple, representing its coordinates) is a boundary cell.

        :param c: a 2-tuple representing the row- and column-index of the cell

        :return: True if the given cell is a boundary cell, False otherwise
        """
        return c[0] == 0 or c[0] == self._rows - 1 or c[1] == 0 or c[1] == self._cols - 1

    def __call__(self, n, c, t):
        """
        The Sandpile rule to apply.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        if self._is_closed_boundary and self._is_in_boundary(c):
            return 0  # closed boundary conditions

        for grain_addition in self._grain_additions:
            if t == grain_addition.timestep and c == grain_addition.cell_index:
                return n[1][1] + 1

        # this cell's activity is the value of the center of the von Neumann neighbourhood
        current_activity = n[1][1]
        new_activity = current_activity

        # this assumes a von Neumann neighbourhood of radius 1
        neighbour_activities = [n[0][1], n[1][0], n[1][2], n[2][1]]

        for neighbour_activity in neighbour_activities:

            if neighbour_activity >= self._K:
                new_activity += 1

        if current_activity >= self._K:
            new_activity -= self._K

        return new_activity


class _GrainAddition:
    """
    A representation of the addition of a grain of sand to the sandpile.
    """
    def __init__(self, cell_index, timestep):
        """
        Create an instance of a `_GrainAddition`.

        :param cell_index: the index of the cell that will have a grain of sand added

        :param timestep: the timestep at which the grain addition occurs
        """
        self.cell_index = cell_index
        self.timestep = timestep



"""
CellPyLib
=========

CellPyLib is a library for working with Cellular Automata.

For complete documentation, see: https://cellpylib.org
"""

__version__ = "2.3.1"

from .ca_functions import BaseRule, AsynchronousRule, ReversibleRule, binary_rule, init_simple, nks_rule, \
    totalistic_rule, plot_multiple, bits_to_int, int_to_bits, init_random, plot, evolve, until_fixed_point, NKSRule, \
    BinaryRule, TotalisticRule

from .rule_tables import random_rule_table, table_walk_through, table_rule

from .entropy import mutual_information, average_cell_entropy, average_mutual_information, shannon_entropy, \
    joint_shannon_entropy

from .ca_functions2d import evolve2d, plot2d, plot2d_slice, plot2d_animate, plot2d_spacetime, init_simple2d, \
    init_random2d, game_of_life_rule

from .bien import binary_derivative, cyclic_binary_derivative, ktbien, tbien, bien

from .apen import apen

from .hopfield_net import HopfieldNet

from .ctrbl_rule import CTRBLRule

from .langtons_loop import LangtonsLoop

from .sdsr_loop import SDSRLoop

from .evoloop import Evoloop

from .sandpile import Sandpile



import matplotlib.pyplot as plt
import numpy as np


def plot(ca, title='', *, colormap='Greys', xlabel='', ylabel='time', show=True, **imshow_kwargs):
    """
    Plots the given cellular automaton.

    :param ca: the cellular automaton to plot

    :param title: the title to place on the plot (default is empty)

    :param colormap: the colormap to use (default is 'Greys')

    :param xlabel: the label of the x-axis (default is empty)

    :param ylabel: the label of the y-axis (default 'time')

    :param show: show the plot (default is True)

    :param imshow_kwargs: keyword arguments for the Matplotlib `imshow` function
    """
    cmap = plt.get_cmap(colormap)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.imshow(ca, interpolation='none', cmap=cmap, **imshow_kwargs)
    if show:
        plt.show()


def plot_multiple(ca_list, titles, *, colormap='Greys', xlabel='', ylabel='time', show=True, **imshow_kwargs):
    """
    Plots multiple cellular automata separately.

    :param ca_list: a list of cellular automata

    :param titles: the titles to give the plots; there must be one title for each CA

    :param colormap: the colormap to use for the plots (default is 'Greys')

    :param xlabel: the label of the x-axis (default is empty)

    :param ylabel: the label of the y-axis (default 'time')

    :param show: show the plot (default is True)

    :param imshow_kwargs: keyword arguments for the Matplotlib `imshow` function
    """
    cmap = plt.get_cmap(colormap)
    for i in range(0, len(ca_list)):
        plt.figure(i)
        plt.title(titles[i])
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.imshow(ca_list[i], interpolation='none', cmap=cmap, **imshow_kwargs)
    if show:
        plt.show()


def evolve(cellular_automaton, timesteps, apply_rule, r=1, memoize=False):
    """
    Evolves the given cellular automaton for the specified time steps. Applies the given function to each cell during
    the evolution. A cellular automaton is represented here as an array of arrays, or matrix. This function expects
    an array containing the initial time step (i.e. initial condition, an array) for the cellular automaton. The final
    result is a matrix, where the number of rows equal the number of time steps specified.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step,
                               e.g. [[0,0,0,0,1,0,0,0,0]], or a history of previous states, with the last item in the
                               given array being the starting condition for the evolution,
                               e.g. [[0,0,0,0,0,0,0,0,0], [0,0,0,0,1,0,0,0,0]]

    :param timesteps: the number of time steps in this evolution, or a callable that accepts the cellular automaton
                      (in terms of the history of its evolution) and the current timestep number, and is expected to
                      return a boolean indicating whether the evolution should continue; note that if a number is given,
                      this value refers to the total number of time steps in this cellular automaton evolution, which
                      includes the initial condition

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy array
                       of length 2r + 1 representing the neighbourhood of the cell; the cell identity, which is a scalar
                       representing the index of the cell in the cellular automaton array; the time step, which is a 
                       scalar representing the time step in the evolution

    :param r: the neighbourhood radius; the neighbourhood size will be 2r + 1 (default is 1)

    :param memoize: allowed values are True, False, and "recursive"; if True, then the result of applying the rule on a
                    given neighbourhood will be cached, and used on subsequent invocations of the rule; if "recursive",
                    then a recursive memoized algorithm will be used, in which recursively wider neighbourhoods are
                    cached, along with the result of applying the rule on the cells in the widened neighbourhood; the
                    True and "recursive" options can result in a significant improvement to execution speed if the rule
                    is expensive to invoke; the "recursive" option works best when there are strongly repetitive
                    patterns in the CA, and when the state consists of 2^k cells; if False, then no caching will be
                    used; NOTE: this should only be set to True or "recursive" for rules which do not store any state
                    upon invocation, and for rules which do not depend in the cell index or timestep number (default is
                    False)

    :return: a matrix, containing the results of the evolution, where the number of rows equal the number of time steps
             specified
    """
    if callable(timesteps):
        return _evolve_dynamic(cellular_automaton, timesteps, apply_rule, r, memoize)
    else:
        return _evolve_fixed(cellular_automaton, timesteps, apply_rule, r, memoize)


def _evolve_fixed(cellular_automaton, timesteps, apply_rule, r, memoize):
    """
    Evolves the given cellular automaton for the given number of fixed timesteps.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step,
                               e.g. [[0,0,0,0,1,0,0,0,0]], or a history of previous states, with the last item in the
                               given array being the starting condition for the evolution,
                               e.g. [[0,0,0,0,0,0,0,0,0], [0,0,0,0,1,0,0,0,0]]

    :param timesteps: the number of time steps in this evolution; this value refers to the total number of time steps in
                      this cellular automaton evolution, which includes the initial condition

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy array
                       of length 2r + 1 representing the neighbourhood of the cell; the cell identity, which is a scalar
                       representing the index of the cell in the cellular automaton array; the time step, which is a
                       scalar representing the time step in the evolution

    :param r: the neighbourhood radius; the neighbourhood size will be 2r + 1

    :param memoize: the memoization flag; one of True, False, or "recursive"

    :return: a matrix, containing the results of the evolution, where the number of rows equal the number of time steps
             specified
    """
    initial_conditions = cellular_automaton[-1]
    _, cols = cellular_automaton.shape
    array = np.zeros((timesteps, cols), dtype=cellular_automaton.dtype)
    array[0] = initial_conditions
    cell_indices = list(range(len(initial_conditions)))

    memo_table = {}

    for t in range(1, timesteps):
        cells = array[t - 1]
        strides = _index_strides(np.arange(len(cells)), 2 * r + 1)
        neighbourhoods = cells[strides]
        if memoize is "recursive":
            next_state = np.zeros(len(cells), dtype=cellular_automaton.dtype)
            _step(cell_indices, cells, next_state, memo_table, apply_rule, r, t)
            array[t] = next_state
        elif memoize is True:
            array[t] = np.array([_get_memoized(n, c, t, apply_rule, memo_table) for c, n in enumerate(neighbourhoods)])
        elif memoize is False:
            array[t] = np.array([apply_rule(n, c, t) for c, n in enumerate(neighbourhoods)])
        else:
            raise Exception("unsupported memoization option: %s" % memoize)

    return np.concatenate((cellular_automaton, array[1:]), axis=0)


def _evolve_dynamic(cellular_automaton, timesteps, apply_rule, r, memoize):
    """
    Evolves the given cellular automaton for a dynamic number of timesteps.

    :param cellular_automaton: the cellular automaton starting condition representing the first time step,
                               e.g. [[0,0,0,0,1,0,0,0,0]], or a history of previous states, with the last item in the
                               given array being the starting condition for the evolution,
                               e.g. [[0,0,0,0,0,0,0,0,0], [0,0,0,0,1,0,0,0,0]]

    :param timesteps: a callable that accepts the cellular automaton (in terms of the history of its evolution) and the
                      current timestep number, and is expected to return a boolean indicating whether the evolution
                      should continue

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy array
                       of length 2r + 1 representing the neighbourhood of the cell; the cell identity, which is a scalar
                       representing the index of the cell in the cellular automaton array; the time step, which is a
                       scalar representing the time step in the evolution

    :param r: the neighbourhood radius; the neighbourhood size will be 2r + 1

    :param memoize: the memoization flag; one of True, False, or "recursive"

    :return: a matrix, containing the results of the evolution, where the number of rows equal the number of time steps
             specified
    """
    initial_conditions = cellular_automaton[-1]
    _, cols = cellular_automaton.shape
    array = [initial_conditions]
    cell_indices = list(range(len(initial_conditions)))

    memo_table = {}

    t = 1
    while timesteps(np.array(array), t):
        cells = array[-1]
        strides = _index_strides(np.arange(len(cells)), 2 * r + 1)
        neighbourhoods = cells[strides]
        if memoize is "recursive":
            result = np.zeros(len(cells), dtype=cellular_automaton.dtype)
            _step(cell_indices, cells, result, memo_table, apply_rule, r, t)
        elif memoize is True:
            result = [_get_memoized(n, c, t, apply_rule, memo_table) for c, n in enumerate(neighbourhoods)]
        elif memoize is False:
            result = [apply_rule(n, c, t) for c, n in enumerate(neighbourhoods)]
        else:
            raise Exception("unsupported memoization option: %s" % memoize)
        array.append(np.array(result, dtype=cellular_automaton.dtype))
        t += 1

    return np.concatenate((cellular_automaton, array[1:]), axis=0)


def _step(indices, curr_state, next_state, cache, apply_rule, r, t):
    """
    Perform an update on the given next state using the current state and memoization cache, based on
    an even split of the cell indices.

    :param indices: a list of the cell indices of the cells to update

    :param curr_state: the current state (i.e. the state after the previous timestep)

    :param next_state: the next state (i.e. the result after the current timestep)

    :param cache: a dictionary that maps state neighbourhoods to their activities

    :param apply_rule: the rule to apply during each cell update

    :param r: the neighbourhood radius

    :param t: the current timestep
    """
    mid = len(indices) // 2
    left_indices = indices[:mid]
    right_indices = indices[mid:]
    if len(left_indices) > 0:
        _update_state(left_indices, curr_state, next_state, cache, apply_rule, r, t)
    if len(right_indices) > 0:
        _update_state(right_indices, curr_state, next_state, cache, apply_rule, r, t)


def _update_state(indices, curr_state, next_state, cache, apply_rule, r, t):
    """
    Perform an update on the given next state using the current state and memoization cache.

    :param indices: a list of the cell indices of the cells to update

    :param curr_state: the current state (i.e. the state after the previous timestep)

    :param next_state: the next state (i.e. the result after the current timestep)

    :param cache: a dictionary that maps state neighbourhoods to their activities

    :param apply_rule: the rule to apply during each cell update

    :param r: the neighbourhood radius

    :param t: the current timestep
    """
    # get the state string for the state given by the indices
    start = indices[0]
    end = indices[-1]
    neighbourhood_indices = range(start - r, end + 1 + r)
    neighbourhood = curr_state.take(neighbourhood_indices, mode='wrap')
    state_string = neighbourhood.tobytes()

    if state_string in cache:
        # update next_state with next vals from cache
        next_state[indices] = cache[state_string]
    else:
        if len(indices) > 1:
            _step(indices, curr_state, next_state, cache, apply_rule, r, t)
        else:
            # invoke rule and update next_state for cell
            val = apply_rule(neighbourhood, start, t)
            next_state[start] = val
        # get the result from the next_state for the left_indices and place in cache
        vals_to_cache = next_state[indices]
        cache[state_string] = vals_to_cache


def _get_memoized(n, c, t, apply_rule, memoization_table):
    """
    Checks if the result of `apply_rule` is in the memoization table according to the neighbourhood, `n`,
    and returns the associated value if it is. Otherwise, `apply_rule` is invoked and the result is stored in
    the memoization table, and then returned.

    :param n: a numpy array representing the neighbourhood

    :param c: an int representing the current cell's index

    :param t: an int, representing the current timestep

    :param apply_rule: a function representing the rule to be applied to each cell during the evolution; this function
                       will be given three arguments, in the following order: the neighbourhood, which is a numpy array
                       of length 2r + 1 representing the neighbourhood of the cell; the cell identity, which is a scalar
                       representing the index of the cell in the cellular automaton array; the time step, which is a
                       scalar representing the time step in the evolution

    :param memoization_table: a dictionary mapping a neighbourhood to the rule result for such a neighbourhood

    :return: the result of invoking the given rule with the given n, c, and t arguments
    """
    key = n.tobytes()
    if key in memoization_table:
        return memoization_table[key]
    else:
        result = apply_rule(n, c, t)
        memoization_table[key] = result
        return result


def _index_strides(arr, window_size):
    """
    Returns an array with dimensions len(cells) x window_size, representing the cell indices of the neighbourhood
    of each cell.

    :param arr: an array containing the cell indices; e.g. if there are 5 cells, then the argument
                will be [0, 1, 2, 3, 4]

    :param window_size: the size of the neighbourhood

    :return: an array with dimensions len(cells) x window_size, representing the cell indices of the neighbourhood
             of each cell
    """
    # this function is based on code in http://www.credid.io/cellular-automata-python-2.html
    arr = np.concatenate((arr[-window_size // 2 + 1:], arr, arr[:window_size // 2]))
    shape = arr.shape[:-1] + (arr.shape[-1] - window_size + 1, window_size)
    strides = arr.strides + (arr.strides[-1],)
    return np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)


def bits_to_int(bits):
    """
    Converts a binary array representing a binary number into the corresponding int.

    :param bits: a list of 1s and 0s, representing a binary number

    :return: and int representing the corresponding number
    """
    total = 0
    for shift, j in enumerate(bits[::-1]):
        if j:
            total += 1 << shift
    return total


def int_to_bits(num, num_digits):
    """
    Converts the given number, `num`, to the corresponding binary number in the form of a NumPy array of 1s and 0s
    comprised of `num_digits` digits.

    :param num: the number, in base 10, to convert into binary

    :param num_digits: the number of digits the binary number should contain

    :return: a NumPy array of 1s and 0s representing the corresponding binary number
    """
    converted = list(map(int, bin(num)[2:]))
    return np.pad(converted, (num_digits - len(converted), 0), 'constant')


def binary_rule(neighbourhood, rule, scheme=None, powers_of_two=None):
    """
    Converts the given rule number to a binary representation, and uses this to determine the value to return.
    The process is approximately described as:

    .. code-block:: text

        1. convert state to int, so [1,0,1] -> 5, call this state_int

        2. convert rule to binary, so 254 -> [1,1,1,1,1,1,1,0], call this rule_bin_array

        3. new value is rule_bin_array[7 - state_int]
           we subtract 7 from state_int to be consistent with the numbering scheme used in NKS
           in NKS, rule 254 for a 1D binary cellular automaton is described as:

          [1,1,1]  [1,1,0]  [1,0,1]  [1,0,0]  [0,1,1]  [0,1,0]  [0,0,1]  [0,0,0]
             1        1        1        1        1        1        1        0

    If None is provided for the scheme parameter, the neighbourhoods are listed in lexicographic order (the reverse of
    the NKS convention). If 'nks' is provided for the scheme parameter, the NKS convention is used for listing the 
    neighbourhoods.

    :param neighbourhood: a binary array of length 2r + 1

    :param rule: an int or a binary array indicating the cellular automaton rule number

    :param scheme: can be None (default) or 'nks'; if 'nks' is given, the rule numbering scheme used in NKS is used

    :param powers_of_two: a pre-computed array containing the powers of two, e.g. [4,2,1]; can be None (default) or an
                          array of length len(neighbourhood); if an array is given, it will used to speed up the
                          calculation of state_int

    :return: the result, 0 or 1, of applying the given rule on the given state
    """
    if powers_of_two is None:
        state_int = bits_to_int(neighbourhood)
    else:
        assert len(powers_of_two) == len(neighbourhood)
        state_int = neighbourhood.dot(powers_of_two)
    n = 2 ** len(neighbourhood)
    if isinstance(rule, (list, np.ndarray)):
        assert len(rule) == n
        rule_bin_array = rule
    else:
        rule_bin_array = int_to_bits(rule, n)
    if scheme == 'nks':
        return rule_bin_array[(n-1) - state_int]
    return rule_bin_array[state_int]


def nks_rule(neighbourhood, rule):
    """
    A convenience function, that calls binary_rule with scheme = 'nks'.

    :param neighbourhood: a binary array of length 2r + 1

    :param rule: an int indicating the cellular automaton rule number

    :return: the result, 0 or 1, of applying the given rule on the given state
    """
    return binary_rule(neighbourhood, rule, scheme='nks')


def totalistic_rule(neighbourhood, k, rule):
    """
    The totalistic rule as described in NKS. The average color is mapped to a whole number in [0, k - 1].
    The rule number is in base 10, but interpreted in base k. For a 1-dimensional cellular automaton, there are
    3k - 2 possible average colors in the 3-cell neighbourhood. There are n(k - 1) + 1 possible average colors for a 
    k-color cellular automaton with an n-cell neighbourhood.

    :param neighbourhood: a k-color array of any size

    :param k: the number of colors in this cellular automaton, where only 2 <= k <= 36 is supported

    :param rule: the k-color cellular automaton rule number in base 10, interpreted in base k

    :return: the result, a number from 0 to k - 1, of applying the given rule on the given state
    """
    # e.g. np.base_repr(777, base=3) -> '1001210'; the zfill pads the string with zeroes: '1'.zfill(3) -> '001'
    #   Bases greater than 36 not handled in base_repr.
    n = neighbourhood.size
    rule_string = np.base_repr(rule, base=k).zfill(n*(k - 1) + 1)
    if len(rule_string) > n*(k - 1) + 1:
        raise ValueError("rule number out of range")
    neighbourhood_sum = np.sum(neighbourhood)
    # the rightmost element of the rule is for the average color 0, in NKS convention
    return int(rule_string[n*(k - 1) - neighbourhood_sum], k)


def init_simple(size, val=1, dtype=np.int32):
    """
    Returns an array initialized with zeroes, with its center value set to the specified value, or 1 by default.

    :param size: the size of the array to be created

    :param val: the value to be used in the center of the array (1, by default)

    :param dtype: the data type

    :return: a vector with shape (1, size), with its center value initialized to the specified value, or 1 by default
    """
    x = np.zeros(size, dtype=dtype)
    x[len(x)//2] = val
    return np.array([x])


def init_random(size, k=2, n_randomized=None, empty_value=0, dtype=np.int32):
    """
    Returns a randomly initialized array with values consisting of numbers in {0,...,k - 1}, where k = 2 by default.
    If dtype is not an integer type, then values will be uniformly distributed over the half-open interval [0, k - 1).

    :param size: the size of the array to be created

    :param k: the number of states in the cellular automaton (2, by default)

    :param n_randomized: the number of randomized sites in the array; this value must be >= 0 and <= size, if specified;
                         if this value is not specified, all sites in the array will be randomized; the randomized sites
                         will be centered in the array, while all others will have an empty value

    :param empty_value: the value to use for non-randomized sites (0, by default)

    :param dtype: the data type

    :return: a vector with shape (1, size), randomly initialized with numbers in {0,...,k - 1}
    """
    if n_randomized is None:
        n_randomized = size
    if n_randomized > size or n_randomized < 0:
        raise ValueError("the number of randomized sites, if specified, must be >= 0 and <= size")
    pad_left = (size - n_randomized) // 2
    pad_right = (size - n_randomized) - pad_left
    if np.issubdtype(dtype, np.integer):
        rand_nums = np.random.randint(k, size=n_randomized, dtype=dtype)
    else:
        rand_nums = np.random.uniform(0, k - 1, size=n_randomized).astype(dtype)
    return np.array([np.pad(np.array(rand_nums), (pad_left, pad_right), 'constant', constant_values=empty_value)])


def until_fixed_point():
    """
    Returns a callable to be used as the `timesteps` argument to the `evolve` and `evolve2d` functions, that will
    result in the evolution being halted when there have been no changes to the state of the CA in the
    last timestep. That is, if the last state of the CA is the same as the second-to-last state, the
    callable will return `False`, and evolution will be halted.

    :return: a callable to be used as the `timesteps` argument to the `evolve` and `evolve2d` functions
    """
    def _timesteps(ca, t):
        if len(ca) > 1:
            return False if (ca[-2] == ca[-1]).all() else True
        return True
    return _timesteps


class BaseRule:
    """
    A base rule class for custom rules to extend. A rule is a callable that accepts three parameters:
    1, the current cell's neighbourhood; 2, the index identifying the current cell; 3, an int identifying the
    current timestep. The rule returns the activity of the current cell at the next timestep.
    """
    def __call__(self, n, c, t):
        """
        The rule to be implemented by subclasses.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        raise NotImplementedError


class NKSRule(BaseRule):
    """
    An Elementary Cellular Automaton rule, indexed according the scheme in NKS.
    """
    def __init__(self, nks_rule_number):
        """
        Creates an instance of an NKS rule.

        :param nks_rule_number: an int indicating the cellular automaton rule number
        """
        self._nks_rule_number = nks_rule_number

    def __call__(self, n, c, t):
        """
        The NKS rule to apply.

        :param n: a binary array of length 2r + 1

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the result, 0 or 1, of applying the given rule on the given state
        """
        return nks_rule(n, self._nks_rule_number)


class BinaryRule(BaseRule):
    """
    A binary representation of the given rule number, which is used to determine the value to return.
    The process is approximately described as:

    .. code-block:: text

        1. convert state to int, so [1,0,1] -> 5, call this state_int

        2. convert rule to binary, so 254 -> [1,1,1,1,1,1,1,0], call this rule_bin_array

        3. new value is rule_bin_array[7 - state_int]
           we subtract 7 from state_int to be consistent with the numbering scheme used in NKS
           in NKS, rule 254 for a 1D binary cellular automaton is described as:

          [1,1,1]  [1,1,0]  [1,0,1]  [1,0,0]  [0,1,1]  [0,1,0]  [0,0,1]  [0,0,0]
             1        1        1        1        1        1        1        0

    If None is provided for the scheme parameter, the neighbourhoods are listed in lexicographic order (the reverse of
    the NKS convention). If 'nks' is provided for the scheme parameter, the NKS convention is used for listing the
    neighbourhoods.
    """
    def __init__(self, rule, scheme=None, powers_of_two=None):
        """
        Creates an instance of a binary rule.

        :param rule: an int or a binary array indicating the cellular automaton rule number

        :param scheme: can be None (default) or 'nks'; if 'nks' is given, the rule numbering scheme used in NKS is used

        :param powers_of_two: a pre-computed array containing the powers of two, e.g. [4,2,1]; can be None (default) or
                              an array of length len(neighbourhood); if an array is given, it will used to speed up the
                              calculation of state_int
        """
        self._rule = rule
        self._scheme = scheme
        self._powers_of_two = powers_of_two

    def __call__(self, n, c, t):
        """
        The binary rule to apply.

        :param n: a binary array of length 2r + 1

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the result, 0 or 1, of applying the given rule on the given state
        """
        return binary_rule(n, self._rule, self._scheme, self._powers_of_two)


class TotalisticRule(BaseRule):
    """
    The totalistic rule as described in NKS. The average color is mapped to a whole number in [0, k - 1].
    The rule number is in base 10, but interpreted in base k. For a 1-dimensional cellular automaton, there are
    3k - 2 possible average colors in the 3-cell neighbourhood. There are n(k - 1) + 1 possible average colors for a
    k-color cellular automaton with an n-cell neighbourhood.
    """
    def __init__(self, k, rule):
        """
        Creates an instance of a totalistic rule.

        :param k: the number of colors in this cellular automaton, where only 2 <= k <= 36 is supported

        :param rule: the k-color cellular automaton rule number in base 10, interpreted in base k
        """
        self._k = k
        self._rule = rule

    def __call__(self, n, c, t):
        """
        The totalistic rule to apply.

        :param n: a k-color array of any size

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the result, a number from 0 to k - 1, of applying the given rule on the given state
        """
        return totalistic_rule(n, self._k, self._rule)


class ReversibleRule(BaseRule):
    """
    An elementary cellular automaton rule explicitly set up to be reversible.
    """
    def __init__(self, init_state, rule_number):
        """
        Creates a reversible elementary cellular automata rule by taking into consideration the previous state of a 
        cell, by taking the XOR of the rule's normal output with the previous state to get the new state.

        :param init_state: a vector representing the initial previous state of the cells, consisting of binary values

        :param rule_number: the elementary cellular automata rule number to be used, in NKS convention
        """
        self._previous_state = init_state
        self._rule_number = rule_number

    def __call__(self, n, c, t):
        """
        The reversible rule to apply.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        regular_result = nks_rule(n, self._rule_number)
        new_result = regular_result ^ self._previous_state[c]
        self._previous_state[c] = n[len(n) // 2]
        return new_result


class AsynchronousRule(BaseRule):
    """
    Creates an asynchronous cellular automaton rule with a cyclic update scheme. Also known as a sequential cellular
    automaton rule, in NKS. This rule wraps a given rule, making the given rule asynchronous. This rule works for 
    both 1D and 2D cellular automata.

    This rule requires the specification of an update order (if none is provided, then an order is constructed based
    on the number of cells in the CA). An update order specifies which cell will be updated as the CA evolves. For
    example, the update order [2, 3, 1] states that cell 2 will be updated in the next timestep, followed by cell 3 in
    the subsequent timestep, and then cell 1 in the timestep after that. This update order is adhered to for the
    entire evolution of the CA. Cells that are not being updated do not have the rule applied to them in that timestep.

    An option is provided to randomize the update order at the end of each cycle (i.e. timestep). This is equivalent to
    selecting a cell randomly at each timestep to update, leaving all others unchanged during that timestep.
    """
    def __init__(self, apply_rule, update_order=None, num_cells=None, randomize_each_cycle=False):
        """
        Constructs an asynchronous rule out of a given rule. Either the update_order or num_cells parameter must be
        specified. If no update_order is given, then the num_cells parameter must be specified, and an update order
        list will be constructed and shuffled.

        :param apply_rule: the rule that will be made asynchronous

        :param update_order: a list containing the indices of the cells in the CA, specifying the update order; if the
                             CA is 2D, then instead of indices, cell coordinates are expected (e.g. ((0,1), (2,3),...))

        :param num_cells: an int specifying the total number of cells in the CA if it is 1D, or a 2-tuple representing
                          the height and width of the CA if it is 2D

        :param randomize_each_cycle: whether to shuffle the update order list after each complete cycle
        """
        if update_order is None and num_cells is None:
            raise ValueError("either update_order or num_cells must be specified")
        self._apply_rule = apply_rule
        if update_order is not None:
            self._update_order = update_order
        else:
            self._init_update_order(num_cells)
            self._shuffle_update_order()
        self._curr = 0
        self._num_applied = 0
        self._randomize_each_cycle = randomize_each_cycle

    def _init_update_order(self, num_cells):
        if isinstance(num_cells, tuple):
            self._update_order = [(i, j) for i in range(num_cells[0]) for j in range(num_cells[1])]
        elif isinstance(num_cells, int):
            self._update_order = np.arange(num_cells)
        else:
            raise TypeError("num_cells must be either an int (1D CA) or a 2-tuple (2D CA)")

    def _shuffle_update_order(self):
        np.random.shuffle(self._update_order)

    def __call__(self, n, c, t):
        """
        The asynchronous rule to apply.

        :param n: the neighbourhood

        :param c: the index of the current cell

        :param t: the current timestep

        :return: the activity of the current cell at the next timestep
        """
        if self._in_update_order(c, n):
            self._num_applied += 1
        if not self._should_update(c, n):
            self._check_for_end_of_cycle()
            return self._current_cell_value(n)
        self._check_for_end_of_cycle()
        return self._apply_rule(n, c, t)

    def _in_update_order(self, c, n):
        return c in self._update_order

    def _should_update(self, c, n):
        return c == self._update_order[self._curr]

    def _check_for_end_of_cycle(self):
        if self._num_applied == len(self._update_order):
            self._curr = (self._curr + 1) % len(self._update_order)
            self._num_applied = 0
            if self._randomize_each_cycle:
                self._shuffle_update_order()

    def _current_cell_value(self, n):
        if len(n.shape) == 1:
            return n[len(n)//2]
        elif len(n.shape) == 2:
            return n[n.shape[0]//2][n.shape[1]//2]
        else:
            raise TypeError("unexpected neighbourhood dimensions: %s" % n.shape)



from setuptools import setup
import re

INIT_FILE = "cellpylib/__init__.py"

with open(INIT_FILE) as fid:
    file_contents = fid.read()
    match = re.search(r"^__version__\s?=\s?['\"]([^'\"]*)['\"]", file_contents, re.M)
    if match:
        version = match.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s" % INIT_FILE)

setup(name="cellpylib",
      version=version,
      description="CellPyLib, A library for working with Cellular Automata, for Python.",
      long_description="CellPyLib is a library for working with Cellular Automata, for Python. "
                       "Currently, only 1- and 2-dimensional k-color cellular automata with "
                       "periodic boundary conditions are supported. The size of the "
                       "neighbourhood can be adjusted. ",
      license="Apache License 2.0",
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.6',
      ],
      url='http://github.com/lantunes/cellpylib',
      author="Luis M. Antunes",
      author_email="lantunes@gmail.com",
      packages=["cellpylib"],
      keywords=["cellular automata", "complexity", "complex systems", "computation", "non-linear dynamics"],
      python_requires='>3.6',
      install_requires=["numpy >= 1.15.4", "matplotlib >= 3.0.2"])



import math

import numpy as np


def shannon_entropy(string):
    """
    Calculates the Shannon entropy for the given string.

    :param string: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the Shannon entropy, in bits
    """
    symbols = dict.fromkeys(list(string))
    symbol_probabilities = [float(string.count(symbol)) / len(string) for symbol in symbols]
    H = -sum([p_symbol * math.log(p_symbol, 2.0) for p_symbol in symbol_probabilities])
    return H + 0  # add 0 as a workaround so we don't end up with -0.0


def average_cell_entropy(cellular_automaton):
    """
    Calculates the average cell entropy in the given cellular automaton, where entropy is the Shannon entropy.
    In the case of a 1D cellular automaton, the state of a cell over time is represented as a string, and its entropy
    is calculated. The same is done for all cells in this cellular automaton, and the average entropy is returned.

    :param cellular_automaton: the cellular automaton to perform this operation on

    :return: a real number representing the average cell Shannon entropy, in bits
    """
    num_cols = cellular_automaton.shape[1]
    entropies = []
    for i in range(0, num_cols):
        cell_states_over_time = ''.join([str(x) for x in cellular_automaton[:, i]])
        entropy = shannon_entropy(cell_states_over_time)
        entropies.append(entropy)
    return np.mean(entropies)


def joint_shannon_entropy(stringX, stringY):
    """
    Calculates the joint Shannon entropy between the given strings, which must be of the same length.

    :param stringX: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :param stringY: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the joint Shannon entropy between the given strings, in bits
    """
    X = np.array(list(stringX))
    Y = np.array(list(stringY))
    joint_symbol_probabilities = []
    for x in set(X):
        for y in set(Y):
            joint_symbol_probabilities.append(np.mean(np.logical_and(X == x, Y == y)))
    return sum([-p * np.log2(p) for p in joint_symbol_probabilities if p != 0])


def mutual_information(stringX, stringY):
    """
    Calculates the mutual information between the given strings, which must be of the same length.

    :param stringX: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :param stringY: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the mutual information between the given strings, in bits
    """
    return shannon_entropy(stringX) + shannon_entropy(stringY) - joint_shannon_entropy(stringX, stringY)


def average_mutual_information(cellular_automaton, temporal_distance=1):
    """
    Calculates the average mutual information between a cell and itself at the next n time steps, given by the 
    specified temporal distance. A temporal distance of 1 means the next time step.

    For example, consider the following string, '00101010110', which represents the state of a cell over 11 time steps.
    The strings which will be used for the computation of the mutual information between a cell and itself at the
    next time step are: '0010101011' and '0101010110', since we pair each time-step value with its next value:

    .. code-block:: text

        " 00101010110"
        "00101010110 "

    :param cellular_automaton: the cellular automaton to perform this operation on

    :param temporal_distance: the size of temporal separation, where the value must be greater than 0 and
                              less than the number of time steps.

    :return: a real number representing the average mutual information between a cell and itself at the next time step,
             in bits
    """
    num_cols = cellular_automaton.shape[1]
    if not (0 < temporal_distance < num_cols):
        raise ValueError("the temporal distance must be greater than 0 and less than the number of time steps")
    mutual_informations = []
    for i in range(0, num_cols):
        cell_states_over_time = ''.join([str(x) for x in cellular_automaton[:, i]])
        mi = mutual_information(cell_states_over_time[:-temporal_distance], cell_states_over_time[temporal_distance:])
        mutual_informations.append(mi)
    return np.mean(mutual_informations)
